from __future__ import print_function
#  Name   :fig_axes
#
#          this class to manage matplotlib.axes
#          the role of this class is similar to
#          fig_section in iScope
#
#
#  Author :
#         Syun'ichi Shiraiwa
#  E-mail :
#         shiraiwa@psfc.mit.edu
#
#
#  History :
#         2012. 04 first version
#         2013. 01 completing cbar....
#
# *******************************************
#     Copyright(c) 2012- PSFC, MIT
# *******************************************
import os
import matplotlib
import weakref
import six
import wx
import numpy as np
from matplotlib.lines import Line2D
from matplotlib.colors import Colormap
import matplotlib.transforms as transforms
from scipy.interpolate import griddata
import matplotlib.ticker as mticker


import ifigure

from ifigure.mto.fig_obj import FigObj
from ifigure.mto.fig_plot import FigPlot
from ifigure.mto.fig_contour import FigContour
from ifigure.mto.fig_image import FigImage
from ifigure.mto.axis_param import AxisXParam, AxisYParam, AxisZParam, AxisCParam
from ifigure.widgets.axes_range_subs import (AdjustableRangeHolderCbar,
                                             AdjustableRangeHolder)

from ifigure.widgets.canvas.file_structure import *
import ifigure.widgets.canvas.custom_picker as cpicker

import ifigure.events
from ifigure.utils.geom import scale_rect

import ifigure.utils.pickle_wrapper as pickle
import ifigure.utils.cbook as cbook

from ifigure.ifigure_config import isMPL33

from ifigure.widgets.undo_redo_history import (GlobalHistory,
                                               UndoRedoArtistProperty,
                                               UndoRedoFigobjProperty,
                                               UndoRedoFigobjMethod,
                                               UndoRedoAddRemoveArtists)

from ifigure.utils.edit_list import DialogEditListTab, DialogEditList

from ifigure.utils.args_parser import ArgsParser
import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('FigAxes')

#color_cycle = matplotlib.rcParams['axes.color_cycle']

#from mpl_toolkits.axes_grid1.inset_locator import inset_axes

unique_label = 0


class FigAxes(FigObj,  AdjustableRangeHolder):
    def __init__(self, area=[0., 0., 1., 1.],
                 margin=[0.15, 0.15, 0.15, 0.15],
                 src=None, *args, **kywds):
        #        self.child=[]
        super(FigAxes, self).__init__(*args, **kywds)
        self.setp("margin", margin)
        self.setp("use_def_margin", True)
        self.setp("area", area)  # left bottom w h
        self._generic_axes = True
        self._margin_bk = [margin, True]
        self._nomargin_mode = False
        self._xaxis = []
        self._yaxis = []
        self._zaxis = []
        self._caxis = []
        self._rect = None
        self._3D = False
        self._3d_pane_color = self._3d_pane_var(
            ('lightgrey', 'lightgrey', 'lightgrey'),
            (1, 1, 1))
        self._use_gl = True
        self._extra_margin = [0., 0., 0., 0.]  # for edge only axis
        # flag to suppress tick label except when
        # an axes is the edge-most axes
        # top, bottom, left, right
        self._edge_only = [False, False, False, False]
        self.add_axis_param(dir='x')
        self.add_axis_param(dir='y')
        self._hastwin = [False, False]

        self.setp("xminortickmode", ['off', 'off'])
        self.setp("yminortickmode", ['off', 'off'])
        self.setp("grid", [False, False, False])  # x, y, z
        self.setp("aspect", 'auto')
        self.setp("show_axis", True)
        self.setp("use_def_size", [True, True, True])  # default text size
        self.setp("title_labelinfo",  ['', 'black', 'default',
                                       'default', 'default', 'default'])
        self.setp('axis_bgalpha', 1.0)
        self.setp('axis_bglinewidth', 0.0)
        self.setp('axis_bglinestyle', 'solid')
        self.setp('axis_bgedgecolor', (0, 0, 0, 0))
        self.setp('axis_bgfacecolor', (1, 1, 1, 1))

    @classmethod
    def isFigAxes(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'axes'

    @classmethod
    def property_in_file(self):
        from ifigure.ifigure_config import isMPL2
        if isMPL2:
            return ["frame_on",
                    "lighting", "azim",
                    "elev", "dist", "_upvec", "_use_clip",
                    "_use_frustum", "_show_3d_axes"]
        else:
            if self._3D:
                return ["lighting", "azim",
                        "elev", "dist", "_upvec", "_use_clip",
                        "_use_frustum", "_show_3d_axes"]
            else:
                return ["frame_on",  # "axis_bgcolor",
                        "lighting", "azim",
                        "elev", "dist", "_upvec", "_use_clip",
                        "_use_frustum", "_show_3d_axes"]

    @classmethod  # define _attr values to be saved
    def attr_in_file(self):
        #       fig_axes store all getp() without this modification
        return []

    @property
    def ispickable_a(self):
        return False

    def can_have_child(self, child=None):
        if child is None:
            return True
        from ifigure.mto.fig_book import FigBook
        from ifigure.mto.fig_page import FigPage
        if isinstance(child, FigPage):
            return False
        if isinstance(child, FigBook):
            return False
        if (isinstance(child, FigAxes) and
                not isinstance(child, FigInsetAxes)):
            return False
        return True

    def add_axis_param(self, dir='x'):
        attr = '_'+dir+'axis'

        g = getattr(self, attr)
        if len(g) == 0:
            name = dir
        else:
            name = dir+str(len(g)+1)
        if dir == 'x':
            p = AxisXParam(name)
            self._xaxis.append(p)
        elif dir == 'y':
            p = AxisYParam(name)
            self._yaxis.append(p)
        elif dir == 'z':
            p = AxisZParam(name)
            self._zaxis.append(p)
        elif dir == 'c':
            p = AxisCParam(name)
            self._caxis.append(p)
        return p

    def get_axis_param(self, name):
        for x in (self._xaxis + self._yaxis
                  + self._zaxis + self._caxis):
            if x.name == name:
                return x

    def get_axis_param_container_idx(self, idx):
        params = {}
        for x in self._xaxis:
            if idx in x._ax_idx:
                params['x'] = x
                break
        for x in self._yaxis:
            if idx in x._ax_idx:
                params['y'] = x
                break
        for x in self._zaxis:
            if idx in x._ax_idx:
                params['z'] = x
                break
        for x in self._caxis:
            if idx in x._ax_idx:
                params['c'] = x
                break
        return params

    def get_axes_artist_by_name(self, name):
        def _search(l, name):
            for p in l:
                if p.name == name:
                    return p._ax_idx
            return []
        if name[0] == 'x':
            idx = _search(self._xaxis, name)
        elif name[0] == 'y':
            idx = _search(self._yaxis, name)
        elif name[0] == 'z':
            idx = _search(self._zaxis, name)
        elif name[0] == 'c':
            idx = _search(self._caxis, name)
        a = [self._artists[k] for k in idx
             if k < len(self._artists)]
        if len(a) != 0:
            return a
        else:
            return [self._artists[0]]

    def property_in_palette_axes(self):
        #        names = [x.name for x in ( self._xaxis + self._yaxis
        #                                  +self._zaxis + self._caxis)]
        tab = ["common"]
        if self._3D:
            item = [["title", "aspect", "axis",  # frameon is not used in 3D
                     "axis_bgcolor", "axis_bgalpha",
                     "axis3d_bgcolor", "axis3d_bgalpha"], ]
        else:
            item = [["title", "aspect", "frame", "axis",
                     "axis_bgedgenone",
                     "axis_bgcolor",
                     "axis_bgedgecolor",
                     "axis_bglinewidth", "axis_bglinestyle", "axis_bgalpha", ]]
        for i in self._xaxis:
            tab.append(i.name)
            item.append(["axlabel", "axrangeparam13",
                         "axformat",
                         "axlotsize",
                         "axtlcolor", "axticks"])
            if not self._3D:
                item[-1].append("axxpos")
        for i in self._yaxis:
            tab.append(i.name)
            item.append(["axlabel", "axrangeparam13",
                         "axformat",
                         "axlotsize",
                         "axtlcolor", "axticks"])
            if not self._3D:
                item[-1].append("axypos")
        if self._3D:
            for i in self._zaxis:
                tab.append(i.name)
                item.append(["axlabel", "axrangeparam13",
                             "axformat",
                             #                            "axlsize",
                             "axlotsize",
                             "axtlcolor", "axticks"])
        for i in self._caxis:
            tab.append(i.name)
            item.append(["axlabel", "axrangeparam13o",
                         "cmap3", "cxclipunder", "cxclipover"])

        if self._3D:
            if self._use_gl:
                tab.append('gl')
                item.append(["gl_view", "gl_lighting", ])
            else:
                pass

        return tab, item

    def property_for_shell(self):
        tab, item = self.property_in_palette_axes()
        tags = [x for x in tab]
        tags[0] = ''
        if tags[-1] == 'gl':
            tags[-1] = ''
        return tab, item, tags

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'axes.png')
        return [idx1]

    def set_parent(self, parent):
        super(FigAxes, self).set_parent(parent)
        from ifigure.mto.fig_page import FigPage
        if isinstance(self._parent, FigPage):
            if self._parent.get_nomargin():
                self.apply_nomargin_mode()

    def get_container(self, *args, **kywds):
        return super(FigAxes, self).get_container()

    def generate_artist(self, *args, **kywds):
        # this method generate artist
        # if artist does exist, update artist
        # based on the information specifically
        # managed by fig_obj tree. Any property
        # internally managed by matplotlib
        # does not change

        container = self.get_container()
        rect, use_def, margin = self.calc_rect()
        if self.isempty():
            #           print "generating axes"
            from ifigure.matplotlib_mod.axes_mod import AxesMod
            from ifigure.matplotlib_mod.axes3d_mod import Axes3DMod

            if self._3D:
                kywds['use_gl'] = self._use_gl
                ax = Axes3DMod(container, rect, *(self._attr["args"]), **kywds)
                if not ax in container.axes:
                    a = container.add_axes(ax)
                ax.disable_mouse_rotation()
            else:
                ax = AxesMod(container, rect, *(self._attr["args"]),
                             **kywds)
                a = container.add_axes(ax)
#           self.backup_margin_param()

            if container.axes.count(ax) == 0:
                print('error in generating axes...')

            self._artists.append(ax)

            ppp = len(self._artists)-1
            params = self.get_axis_param_container_idx(ppp)
            if not "x" in params:
                self._xaxis[-1].add_ax_idx(ppp)
            if not "y" in params:
                self._yaxis[-1].add_ax_idx(ppp)
            if self._3D:
                if not "z" in params:
                    self._zaxis[-1].add_ax_idx(ppp)
                code = self.get_3d_pane_colorcode()
                self._set_3d_pane_color(code)

#           for name in ('x', 'y'):
#               p = self.get_axis_param(name)

            for artist in self._artists:
                artist.figobj = self
                artist.figobj_hl = []
                artist.set_zorder(self.getp('zorder'))
            lp = self.getp("loaded_property")

            if lp is not None:
                for k, artist in enumerate(self._artists):
                    self.set_artist_property(artist, lp[k])
                    # print lp[k]
                self.delp("loaded_property")

            self.set_title(self.getp('title_labelinfo'), self._artists[0])
            self.set_axis_bgedgecolor(self.getp('axis_bgedgecolor'), None)
            self.set_axis_bgfacecolor(self.getp('axis_bgfacecolor'))
            self.set_axis_bgalpha(self.getp('axis_bgalpha'), None)
            self.set_axis_bglinestyle(self.getp('axis_bglinestyle'), None)
            self.set_axis_bglinewidth(self.getp('axis_bglinewidth'), None)
            self.set_axis_onoff(self.getp('show_axis'), None)

            if self._hastwin[0]:
                self.twinx(new_axesparam=False)
            elif self._hastwin[1]:
                self.twiny(new_axesparam=False)

            for idx, a in enumerate(self._artists):
                params = self.get_axis_param_container_idx(idx)
                for key in params:
                    params[key].set_artist_rangeparam(a)
                    params[key].set_label(a)
                    params[key].set_ticks(a)
                    if not self._3D:
                        params[key].set_tickparam(a, self)

            self._mpl_cb = [a.callbacks.connect('xlim_changed',
                                                self.xylim_changed_cb),
                            a.callbacks.connect('ylim_changed',
                                                self.xylim_changed_cb)]

        else:
            for artist in self._artists:
                artist.set_position(rect)
#        print self._artists[0].get_window_extent().get_points()
        self.set_aspect(self.getp('aspect'))
        self.set_grid(self.getp('grid'))
        for a in self._artists:
            if hasattr(a, 'set_nomargin_mode'):
                a.set_nomargin_mode(self._nomargin_mode)

    def generate_cursor(self, evt, idx, target=None):
        '''
        this routine produce line cursor artists
        '''
        if self.get_3d():
            return
        f_page = self.get_figpage()
        if f_page is None:
            return super(FigAxes, self).generate_cursor(evt, idx)
        if evt.xdata is None:
            return
        if len(self._artists) == 0:
            return
        if target is None:
            kk = 0
        else:
            kk = (0 if not target.axes in self._artists
                  else self._artists.index(target.axes))
        ax = self._artists[kk]
        tx = transforms.blended_transform_factory(ax.transData,
                                                  ax.transAxes)
        ty = transforms.blended_transform_factory(ax.transAxes,
                                                  ax.transData)
        box = transforms.Bbox.from_extents(ax.get_window_extent().extents)

        alpha = ifigure._cursor_config["1dalpha"]
        c1 = ifigure._cursor_config["1dcolor1"]
        c2 = ifigure._cursor_config["1dcolor2"]
        w = ifigure._cursor_config["1dthick"]
        if idx == 0:
            self._cursor1 = [Line2D([evt.xdata, evt.xdata], [0, 1],
                                    figure=f_page._artists[0],
                                    transform=tx, linewidth=w,
                                    linestyle='-', color=c1, alpha=alpha,
                                    clip_box=box, clip_on=True,
                                    markerfacecolor='None'),
                             Line2D([0, 1], [evt.ydata, evt.ydata],
                                    figure=f_page._artists[0],
                                    transform=ty, linewidth=w,
                                    linestyle='-', color=c1,
                                    alpha=alpha, clip_box=box, clip_on=True,
                                    markerfacecolor='None')]
            self._cursor1_a = weakref.ref(ax)
        else:
            self._cursor2 = [Line2D([evt.xdata, evt.xdata], [0, 1],
                                    figure=f_page._artists[0],
                                    transform=tx, linewidth=w,
                                    linestyle='-', color=c2,
                                    alpha=alpha, clip_box=box, clip_on=True,
                                    markerfacecolor='None'),
                             Line2D([0, 1], [evt.ydata, evt.ydata],
                                    figure=f_page._artists[0],
                                    transform=ty, linewidth=w,
                                    linestyle='-', color=c2,
                                    alpha=alpha, clip_box=box, clip_on=True,
                                    markerfacecolor='None')]
            self._cursor2_a = weakref.ref(ax)
        self.get_figpage().add_resize_cb(self)

    def reset_cursor_range(self, evt, idx):
        return
        ### evt and idx is not used
        a = np.inf
        b = -np.inf
        for cx in self._caxis:
            a = np.min((a, cx.range[0], cx.range[1]))
            b = np.max((b, cx.range[0], cx.range[1]))
        self._cursor_range = (a, b)
#    def init_cursor_str(self):
#        self._cursor_str = ['','','']

    def erase_cursor(self):
        self._cursor1 = []
        self._cursor2 = []
        self._cursor1_a = None
        self._cursor2_a = None

    def update_cursor(self, evt, idx, mode=0, target=None):
        if evt.xdata is None:
            return
        x = evt.xdata
        y = evt.ydata
        if mode == 0:
            self._update_cursor_line(idx, x, y)
        if mode == 1 and isinstance(target, Line2D):
            k = np.argmin(abs(target.get_xdata() - evt.xdata))
            x = target.get_xdata()[k]
            y = target.get_ydata()[k]
            self._update_cursor_line(idx, x, y)
            return x, y
        if mode == 2 and isinstance(target, Line2D):
            k = np.argmin(abs(target.get_ydata() - evt.ydata))
            x = target.get_xdata()[k]
            y = target.get_ydata()[k]
            self._update_cursor_line(idx, x, y)
            return x, y
        if mode == 3:
            x, y, z = target.figobj.getp(('x', 'y', 'z'))
            if z.size == x.size*y.size:
                X, Y = np.meshgrid(x, y)
                p1 = np.transpose(np.vstack((X.flatten(), Y.flatten())))
                z0 = float(griddata(p1, z.flatten(),
                                    (evt.xdata, evt.ydata),
                                    method='linear'))
                self._cursor_data[idx] = (evt.xdata,
                                          evt.ydata,
                                          z0)
                if idx == 0:
                    self._cursor_range = (z0, self._cursor_range[1])
                else:
                    self._cursor_range = (self._cursor_range[0], z0)
                if self._cursor_range[0] > self._cursor_range[1]:
                    self.erase_cursor()
                    return

                # draw cursor

                z = target.get_array()
                extent = target.get_extent()
                zp = ((z > self._cursor_range[0]).astype(int) *
                      (z < self._cursor_range[1]).astype(int))

                alpha = ifigure._cursor_config["2dalpha"]
                z = np.empty((zp.shape[0], zp.shape[1], 4))
                z[:, :, 0] = 1 - zp
                z[:, :, 1] = 1 - zp
                z[:, :, 2] = 1 - zp
                z[:, :, 3] = zp*alpha
                self._cursor1 = [self._artists[0].imshow(z,
                                                         extent=extent, aspect='auto',
                                                         origin='lower')]
 #               self._cursor1[0].remove()
                self._cursor1_a = weakref.ref(self._artists[0])
            else:
                print('contour coursor for this mode is not implemented')
                print('make funcy obj here')

        if mode == 4:
            '''
            target.figobj should be either FigImage or FigContour
            '''
            x, y, z = target.figobj.getp(('x', 'y', 'z'))
            if z.size == x.size*y.size:
                X, Y = np.meshgrid(x, y)
                p1 = np.transpose(np.vstack((X.flatten(), Y.flatten())))
                z0 = float(griddata(p1, z.flatten(),
                                    (evt.xdata, evt.ydata),
                                    method='linear'))

                self._cursor_data[idx] = (evt.xdata,
                                          evt.ydata,
                                          z0)
                if idx == 0:
                    self._cursor_range = (z0, self._cursor_range[1])
                else:
                    self._cursor_range = (self._cursor_range[0], z0)
                if self._cursor_range[0] > self._cursor_range[1]:
                    self.erase_cursor()
                    return

                alpha = ifigure._cursor_config["2dalpha"]
#                for a in self._cursor1:
#                    a.remove()
                self._cursor1 = self._artists[0].contour(x, y, z,
                                                         self._cursor_range, alpha=alpha).collections[:]
                self._cursor1_a = weakref.ref(self._artists[0])
#                for a in self._cursor1:
#                    a.remove()
            else:
                print('contour coursor for this mode is not implemented')
                print('make funcy obj here')
        if mode == 5:
            from ifigure.widgets.slice_viewer import SliceViewer
            self._update_cursor_line(idx, x, y)
            app = self.get_root_parent().app

            if (ifigure._cursor_book is None or
                    ifigure._cursor_book() is None):
                #print('creating slice viewer')
                book, viewer = app.open_newbook_in_newviewer(
                    SliceViewer)

                ifigure._cursor_book = weakref.ref(book)
            book = ifigure._cursor_book()
            if not book.isOpen:
                if book.num_page() == 0:
                    book.add_page()
                viewer = app.open_book_in_newviewer(SliceViewer,
                                                    book)
            else:
                viewer = app.find_bookviewer(book)

#            data1, data2 = target.figobj.get_slice(x, y)
            data1, data2 = target.figobj.get_slice(evt.x, evt.y)
            if data1 is None:
                return
            if data1 is None:
                return

            viewer.update_curve(idx, x, y, data1, data2)

    def _update_cursor_line(self, idx, x, y):
        if idx == 0:
            if len(self._cursor1) < 2:
                return
            self._cursor1[0].set_xdata([x, x])
            self._cursor1[1].set_ydata([y, y])
        else:
            if len(self._cursor2) < 2:
                return
            self._cursor2[0].set_xdata([x, x])
            self._cursor2[1].set_ydata([y, y])
        self._cursor_data[idx] = (x, y)

    def valid_cursor(self):
        val = []
        if len(self._artists) == 0:
            return val
        if (self._cursor1_a is not None and
                self._cursor1_a() in self._artists):
            val += self._cursor1
        if (self._cursor2_a is not None and
                self._cursor2_a() in self._artists):
            val += self._cursor2
        return val

    '''
    xylim_changed_cb is used to update pageobj reling
    on axes. GenericPointHolder needs to be updated 
    through this message chain.
    '''

    def xylim_changed_cb(self, a):
        self.set_client_update_artist_request()
#    def ylim_changed_cb(self, a):
#        self.set_client_update_artist_request()

    def del_artist(self, artist=None, delall=False):

        if delall:
            alist = self._artists
        else:
            #           alist=[artist]
            alist = artist

        val = []
        for a in self._artists:
            val.append(self.get_artist_property(a))
        if len(val) != 0:
            self.setp("loaded_property", val)

        if len(alist) is len(self._artists):
            for name, child in self.get_children():
                child.del_artist(delall=True)

        if len(alist) != 0:
            self.highlight_artist(False, alist)
            container = self.get_container()
            for a in alist:
                a.clear()
                a.callbacks.disconnect(self._mpl_cb[0])
                a.callbacks.disconnect(self._mpl_cb[1])
                try:
                    container.delaxes(a)
                except:
                    print("can not do delaxes...?")
                a.axes = None
        super(FigAxes, self).del_artist(alist, delall=delall)

    def get_axesartist_extent(self, w, h, a=None, renderer=None, box=None):
        '''
        a is ignored
        '''
        from ifigure.utils.cbook import GetArtistExtent

        def merge_box(box, box2):
            return (min((box[0], box2[0])),
                    min((box[1], box2[1])),
                    max((box[2], box2[2])),
                    max((box[3], box2[3])),)
        if box is None:
            box = (np.inf, np.inf, -np.inf, -np.inf)

        for a in self._artists:
            xa = a.get_xaxis()
            ya = a.get_yaxis()

            alist = [a, a.title]
            alist += xa.get_ticklabels()
            alist.append(xa.get_label())
            alist.append(xa.get_offset_text())
            alist += ya.get_ticklabels()
            alist.append(ya.get_label())
            alist.append(ya.get_offset_text())
            alist.extend(a.artists)
            #renderer = a.figure._cachedRenderer
            renderer = a.figure.canvas.get_renderer()

            for item in alist:
                # skip this case since it fails
                if (isinstance(item, matplotlib.text.Text) and
                        renderer is None):
                    continue
                try:

                    box1 = item.get_window_extent(renderer).extents
                    box = merge_box(box, box1)
                except:
                    dprint1('skipping obj in axesartist_extent:' + str(item))

            for item in a.texts:
                box = GetArtistExtent(item, box=box,  renderer=renderer)

            box = (max((box[0]-2, 0)),
                   max((box[1]-2, 0)),
                   min((w, box[2]+2)),
                   min((h, box[3]+2)))

        return box

    def calc_rect(self, ignore_pagemargin=False):
        #        print self, 'calc_rect'
        #        import traceback
        #        traceback.print_stack()
        if self._attr["use_def_margin"]:
            margin = self._parent.getp("def_margin")
        else:
            margin = self.getp("margin")

        area = self._attr["area"]
        tm = [margin[i] + self._extra_margin[i]
              for i in range(4)]
        rect = [area[0]+tm[0]*area[2],
                area[1]+tm[3]*area[3],
                (1-tm[0]-tm[1])*area[2],
                (1-tm[2]-tm[3])*area[3]]
        if self._nomargin_mode:
            ignore_pagemargin = True
        if not ignore_pagemargin:
            pm = self.get_figpage()._page_margin
            rect = [(1-pm[0]-pm[1])*rect[0]+pm[0],
                    (1-pm[2]-pm[3])*rect[1]+pm[3],
                    rect[2]*(1-pm[0]-pm[1]),
                    rect[3]*(1-pm[2]-pm[3]), ]
            if self.getp('aspect') == 'equal' and not self._3D:
                params = self.get_axis_param_container_idx(0)
                dx = abs(params['x'].range[1] - params['x'].range[0])
                dy = abs(params['y'].range[1] - params['y'].range[0])
                w, h = self.get_figpage().getp('figsize')
                ratio = (w*rect[2])*dy/dx/(h*rect[3])
                if ratio > 1:
                    #                   v = rect[3] / ratio
                    v = rect[3]*h*dx/dy/w
                    rect[0] = rect[0]+(rect[2] - v)/2.
#                   rect[2] = v
                    rect[2] = rect[3]*h*dx/dy/w
                else:
                    #                   v = rect[2] * ratio
                    v = rect[2]*w*dy/dx/h
                    rect[1] = rect[1]+(rect[3] - v)/2.
#                   rect[3] = v
                    rect[3] = rect[2]*w*dy/dx/h
        self._rect = rect
        return rect, self.getp("use_def_margin"), margin

    def get_rect(self, asbbox=False):
        if self._rect is None:
            self.calc_rect()
        return self._rect

    def set_area(self, area, a=None):
        self.setp("area", area[:])
        rect, use_def, margin = self.calc_rect()
        for a in self._artists:
            a.set_position(rect)
        self.set_client_update_artist_request()
        self.set_bmp_update(False)

    def get_area(self, a=None):
        return self.getp("area")[:]

    def set_margin(self, m, a=None):
        self.setp("margin", m[:])
        rect, use_def, margin = self.calc_rect()
        for a in self._artists:
            a.set_position(rect)
        self.set_client_update_artist_request()
        self.set_bmp_update(False)

    def get_margin(self, m, a=None):
        return self.getp("margin")[:]

    def get_grid(self, a=None):
        return self.getp('grid')

    def set_grid(self, value, a=None):
        if len(self._artists) == 0:
            return
        if self._3D:
            return
        a = self._artists[0]

        a.grid(value[0], 'major', 'x')
        a.grid(value[1], 'major', 'y')
        self.setp('grid', list(value))
#        if self._3D:
#            a.grid(value[2], 'major', 'z')

    def set_aspect(self, value, a=None):
        #        for a in self._artists: a.set_aspect(value)
        if len(self._artists) > 0:
            self._artists[0].set_aspect('auto')
        if self._3D:
            self.setp('aspect', 'auto')
            rect, use_def, margin = self.calc_rect()
            self.setp('aspect', str(value))
            if len(self._artists) > 0:
                ax = self._artists[0]
                if value == 'equal':
                    ax._ignore_screen_aspect_ratio = False
                else:
                    ax._ignore_screen_aspect_ratio = True
        else:
            self.setp('aspect', str(value))
            rect, use_def, margin = self.calc_rect()
        for a in self._artists:
            a.set_position(rect)
        self.set_client_update_artist_request()
        self.set_bmp_update(False)
        if str(value) == 'equal' and not self._3D:
            self.get_figpage().add_resize_cb(self)
        else:
            self.get_figpage().rm_resize_cb(self)

    def get_aspect(self, a=None):
        return self.getp('aspect')

    def set_extra_margin(self, m, a=None):
        self._extra_margin = m[:]

    def get_extra_margin(self, a=None):
        return self._extra_margin[:]

    def set_edge_only(self, m, a=None):
        self._edge_only = m

    def get_edge_only(self, a=None):
        return self._edge_only[:]

    def add_plot(self, name=None, *args, **kywds):
        obj = FigPlot(*args, **kywds)
        if name is None:
            name = self.get_next_name(obj.get_namebase())
        return self.add_child(name, obj)

    def add_contour(self, name=None, *args, **kywds):
        obj = FigContour(*args, **kywds)
        if name is None:
            name = self.get_next_name(obj.get_namebase())
        return self.add_child(name, obj)

    def add_image(self, name=None, *args, **kywds):
        obj = FigImage(*args, **kywds)
        if name is None:
            name = self.get_next_name(obj.get_namebase())
        return self.add_child(name, obj)

    def add_insetaxes(self, name=None, *args, **kywds):
        obj = FigInsetAxes(*args, **kywds)
        if name is None:
            name = self.get_next_name(obj.get_namebase())
        return self.add_child(name, obj)

    def add_colorbar(self, name=None,  *args, **kywds):
        '''
        this routine shoud be accessed from caxis_param.show_cbar
        '''
        obj = FigColorBar(*args, **kywds)
        if name is None:
            name = self.get_next_name(obj.get_namebase())
        return self.add_child(name, obj)

    def del_plot(self, fig_plot):
        self.del_figobj(fig_plot)

#
# setter/getter
#
    def set_axis_onoff(self, value, a=None):
        self.setp("show_axis", value)
        for a in self._artists:
            if value:
                a.set_axis_on()
            else:
                a.set_axis_off()

    def get_axis_onoff(self, a):
        return self.getp("show_axis")

    # nomargin
    def apply_nomargin_mode(self):
        if self.get_figpage().get_nomargin():
            self.backup_margin_param()
            self.setp("margin", [0, 0, 0, 0])
            self.setp("use_def_margin", False)
            for a in self._artists:
                if hasattr(a, 'set_nomargin_mode'):
                    a.set_nomargin_mode(True)
            self._nomargin_mode = True
        else:
            self.setp("margin", self._margin_bk[0])
            self.setp("use_def_margin", self._margin_bk[1])
            for a in self._artists:
                if hasattr(a, 'set_nomargin_mode'):
                    a.set_nomargin_mode(False)
            self._nomargin_mode = False
        rect, use_def, margin = self.calc_rect()
        for a in self._artists:
            a.set_position(rect)
        self.set_client_update_artist_request()

    def backup_margin_param(self):
        if not self._nomargin_mode:
            self._margin_bk = (self.getp("margin"),
                               self.getp("use_def_margin"))
    ### range / autorange / linear-log

    def set_axrangeparam13(self, param, a=None):
        '''
        this is called from aritst widgets in bypass history
        mode. history entry needs to be made in this routine

        param is in the format of request  (name, value)
        '''
        book = self.get_figbook()

        app = wx.GetApp().TopWindow
        viewer = app.find_bookviewer(book)
        if viewer is None:
            return
        canvas = viewer.canvas
        requests = {self: [param]}
        if param[1][0]:
            requests[self] = self.compute_new_range(request=requests[self])
        requests = canvas.expand_requests(requests)

        canvas.send_range_action(requests, 'range')

    def set_axrangeparam(self, param, a=None):
        p = self.get_axis_param(param[0])
        param[1][3] == str(param[1][3])
        if param[1][3] == 'log':
            range = param[1][2]
            range0 = range[0] if range[0] > 0 else range[1]*1e-16
#           range0= max((1e-16, range[0]))
            range1 = range[1]
            if range[1] < range0:
                range1 = 2*range0
            param[1][2] = [range0, range1]
        p.set_rangeparam(param[1])

        # update aritst range param
        if self.isempty():
            return
        for a in self.get_axes_artist_by_name(param[0]):
            p.set_artist_rangeparam(a)
        if self.get_aspect() == 'equal':
            rect, use_def, margin = self.calc_rect()
            for a in self._artists:
                a.set_position(rect)
        self.set_bmp_update(False)
#        ifigure.events.SendRangeChangedEvent(self, name=param[0])
#        print 'send range change event'

    def get_axrangeparam13(self, a, name):
        return self.get_axrangeparam(a, name)

    def get_axrangeparam(self, a, name):
        p = self.get_axis_param(name)
        value = p.get_rangeparam()
        return value

    def set_adjustrange(self, *args):
        self.adjust_axes_range()

    def get_adjustrange(self, *args):
        pass

    def call_handle_axes_change(self, n):
        if self.isempty():
            return
        for child in self.walk_tree():
            if child is not self and not child._suppress:
                child.handle_axes_change({'td': self, 'name': n})

    def set_textprop(self, value, a):
        a.set_text(value[0])
        a.set_color(value[1])
        a.set_family(value[2])
        a.set_weight(value[3])
        a.set_style(value[4])
        a.set_size(float(value[5]))

    def get_textprop(self, a):
        return [a.get_text(),
                a.get_color(),
                a.get_family()[0],
                a.get_weight(),
                a.get_style(),
                str(a.get_size())]

    def set_title(self, value, a=None):
        '''
        set/get_title stores textprop
        to _attr

        calling this using string argment is not recommended.
        '''
        figpage = self.get_figpage()
        if figpage is None:
            return

        tinfo = self.getp('title_labelinfo')
        tfont = figpage.getp('title_font')
        tweight = figpage.getp('title_weight')
        tstyle = figpage.getp('title_style')
        tsize = figpage.getp('title_size')

        if isinstance(value, six.string_types):
            # Don't use this!
            # use viewer.title whenever possible.
            dprint1('set_title is called with string argument. not recommendad')
            if a is not None:
                v0 = self.get_textprop(a.title)
                v0[0] = value
                value = v0
                v2 = [x for x in value]
            else:
                value = [value, 'k', 'default', 'default',
                         'default', 'default']
                v2 = [value, 'k', 'default', 'default',
                      'default', 'default']
        else:
            v2 = [x for x in value]
        if str(value[2]) == 'default':
            v2[2] = tfont
        if str(value[3]) == 'default':
            v2[3] = tweight
        if str(value[4]) == 'default':
            v2[4] = tstyle
        if str(value[5]) == 'default':
            v2[5] = tsize

        if a is not None:
            self.set_textprop(v2, a.title)
        self.setp('title_labelinfo', value)

    def get_title(self, a):
        value = self.getp('title_labelinfo')
        return [x for x in value]

    def set_cxclipunder(self, param, a):
        p = self.get_axis_param(param[0])
        p.clip = (param[1], p.clip[1])
        for a in self._artists:
            p.set_artist_rangeparam(a)
        self.set_bmp_update(False)

    def get_cxclipunder(self, a, name):
        p = self.get_axis_param(name)
        return p.clip[0]

    def set_cxclipover(self, param, a):
        p = self.get_axis_param(param[0])
        p.clip = (p.clip[0], param[1])
        for a in self._artists:
            p.set_artist_rangeparam(a)
        self.set_bmp_update(False)

    def get_cxclipover(self, a, name):
        p = self.get_axis_param(name)
        return p.clip[1]

    def set_axticks(self, param, a=None):
        p = self.get_axis_param(param[0])
        p.ticks = param[1]
        ax = self.get_axes_artist_by_name(param[0])[0]
        p.set_ticks(ax)
        self.set_bmp_update(False)

    def get_axticks(self, a=None, name='x'):
        p = self.get_axis_param(name)
        return p.ticks

    def refresh_nticks(self):
        self.set_bmp_update(False)
        for x in self._xaxis:
            #            x.ticks = nxticks
            a = self.get_axes_artist_by_name(x.name)[0]
            x.set_ticks(a)
        for y in self._yaxis:
            #            y.ticks = nyticks
            a = self.get_axes_artist_by_name(y.name)[0]
            y.set_ticks(a)
        if not self.get_3d():
            return
        for z in self._zaxis:
            #            z.ticks = nzticks
            a = self.get_axes_artist_by_name(z.name)[0]
            z.set_ticks(a)

    def set_axlabel(self, param, a):
        p = self.get_axis_param(param[0])
        p.labelinfo = param[1]
        p.set_label(self.get_axes_artist_by_name(param[0])[0])
        self.set_bmp_update(False)

    def get_axlabel(self, a, name):
        p = self.get_axis_param(name)
        return [x for x in p.labelinfo]

    # axis label color
    def set_axlcolor(self, param, a):
        p = self.get_axis_param(param[0])
        p.lcolor = param[1][0]
        p.otcolor = param[1][1]
        p.set_tickparam(self.get_axes_artist_by_name(param[0])[0], self)
        self.set_bmp_update(False)

    def get_axlcolor(self, a, name):
        p = self.get_axis_param(name)
        return p.lcolor, p.otcolor

    # tick color
    def set_axtcolor(self, param, a):
        from ifigure.utils.cbook import isstringlike
        p = self.get_axis_param(param[0])
        if isstringlike(param[1]):
            p.tcolor = str(param[1])
        else:
            p.tcolor = param[1]
        for a in self.get_axes_artist_by_name(param[0]):
            p.set_tickparam(a, self)
        self.set_bmp_update(False)

    def get_axtcolor(self, a, name):
        p = self.get_axis_param(name)
        return p.tcolor

    # tick/label color
    def set_axtlcolor(self, param, a):
        from ifigure.utils.cbook import isstringlike
        p = self.get_axis_param(param[0])
        if isstringlike(param[1]):
            p.tcolor = str(param[1][0])
        else:
            p.tcolor = param[1][0]
        p.lcolor = param[1][1]
        p.otcolor = param[1][2]
        for a in self.get_axes_artist_by_name(param[0]):
            p.set_tickparam(a, self)
        self.set_bmp_update(False)

    def get_axtlcolor(self, a, name):
        p = self.get_axis_param(name)
        return p.tcolor, p.lcolor, p.otcolor

    # label size
    def set_axlsize(self, param, a):
        p = self.get_axis_param(param[0])
        p.lsize = param[1]
        p.set_tickparam(self.get_axes_artist_by_name(param[0])[0], self)
        self.set_bmp_update(False)

    def get_axlsize(self, a, name):
        p = self.get_axis_param(name)
        return p.lsize

    def set_axotsize(self, param, a):
        p = self.get_axis_param(param[0])
        p.otsize = param[1]
        p.set_tickparam(self.get_axes_artist_by_name(param[0])[0], self)
        self.set_bmp_update(False)

    def get_axotsize(self, a, name):
        p = self.get_axis_param(name)
        return p.otsize

    def set_axlotsize(self, param, a):
        p = self.get_axis_param(param[0])
        p.lsize = 'default' if param[1][0] == 'default' else int(param[1][0])
        p.otsize = 'default' if param[1][1] == 'default' else int(param[1][1])
        p.set_tickparam(self.get_axes_artist_by_name(param[0])[0], self)
        self.set_bmp_update(False)

    def get_axlotsize(self, a, name):
        p = self.get_axis_param(name)
        return p.lsize, p.otsize

    def postprocess_mpltext_edit(self):
        '''
        called after text is edited interactively
        this allows to load mpl text data to figobj
        '''
        a = self._artists[0]
        if self._3D:
            plist = (self._xaxis + self._yaxis
                     + self._zaxis + self._caxis)
        else:
            plist = self._xaxis + self._yaxis + self._caxis
        for p in plist:
            a = self.get_axes_artist_by_name(p.name)[0]
            label = p.get_label_from_mpl(a)
            p.labelinfo[0] = label

        # title
        a = self._artists[0]
        value = self.getp('title_labelinfo')
        value[0] = a.title.get_text()

    # format
    def set_axformat(self, param, a):
        p = self.get_axis_param(param[0])
        p.format = param[1]
        p.set_ticks(self.get_axes_artist_by_name(param[0])[0])
        self.set_bmp_update(False)

    def get_axformat(self, a, name):
        p = self.get_axis_param(name)
        return p.format

    def onResize(self, evt):
        self.erase_cursor()
        if self.getp('aspect') == 'equal':
            rect, use_def, margin = self.calc_rect()
            for a in self._artists:
                a.set_position(rect)
            self.set_client_update_artist_request()


#
#   menu in tree viewer
#


    def tree_viewer_menu(self):
        # return MenuString, Handler, MenuImage
        #  m=('Add Plot',     self.onAddPlot, None),
        #      ('Add Contour',  self.onAddContour, None),
        #     ('Add Image',    self.onAddImage, None),
        m = [('Add InsetAxes', self.onAddInsetAxes, None), ]
        if False in [p.has_cbar() for p in self._caxis]:
            m.append(('Add ColorBar', self.onAddColorBar, None),)
        m.append(('---',        None, None))
        if len(self._artists) == 0:
            m.append(('Realize',    self.onRealize, None))

        return m+super(FigAxes, self).tree_viewer_menu()

    def onRealize(self, e):
        self.realize()
        ifigure.events.SendPVDrawRequest(self)
#       self.onForceUpdate(e)

    def onAddPlot(self, e):
        self.add_plot(None, [0, 1], [0, 1])
        self.realize()
        ifigure.events.SendPVAddFigobj(self)
        e.Skip()

    def onAddInsetAxes(self, e=None):
        ichild = self.add_insetaxes()
        child = self.get_child(ichild)
        child.realize()
        ifigure.events.SendPVAddFigobj(self)
        if e is not None:
            e.Skip()

    def onAddInsetAxesCanvas(self, evt):
        '''
        when canvas menu is selected
        evt.GetEventObject() = canvas
        '''
        ichild = self.add_insetaxes()
        child = self.get_child(ichild)
        child.realize()
        ifigure.events.SendChangedEvent(
            self.get_figpage(), evt.GetEventObject())
        ifigure.events.SendCanvasDrawRequest(evt.GetEventObject(),)
        evt.Skip()

    def onAddColorBar(self, e=None):
        '''
        add color bar when called from project tree browser
        not undoable...
        '''
        if False in [p.has_cbar() for p in self._caxis]:
            for k, p in enumerate(self._caxis):
                if not p.has_cbar():
                    p.show_cbar(self, offset=-0.1*k)
        ifigure.events.SendPVAddFigobj(self)
        e.Skip()

    def onAddContour(self, e):
        self.add_contour(None, np.arange(4).reshape(2, 2))
        self.realize()
        ifigure.events.SendPVAddFigobj(self)
        e.Skip()

    def onAddImage(self, e):
        self.add_image(None, np.zeros([2, 2]))
        self.realize()
        ifigure.events.SendPVAddFigobj(self)
        e.Skip()

# save/load
    def save_data(self, fid=None):
        val = self.getp()
        # print "writing axes param", val
        pickle.dump(val, fid)
        super(FigAxes, self).save_data(fid)

    def save_data2(self, data):
        # the first element is version code
        param = {"_edge_only": self._edge_only,
                 "_margin_bk": self._margin_bk,
                 "_nomargin_mode": self._nomargin_mode,
                 "_3D": self._3D,
                 "_extra_margin": self._extra_margin,
                 "_3d_pane_color": self._3d_pane_color,
                 "_hastwin": self._hastwin}

        axisdata = {"xaxis": [x.make_save_data(self) for x in self._xaxis],
                    "yaxis": [x.make_save_data(self) for x in self._yaxis],
                    "zaxis": [x.make_save_data(self) for x in self._zaxis],
                    "caxis": [x.make_save_data(self) for x in self._caxis]}

        data['FigAxes'] = (1, self.getp(), param, axisdata)
        data = super(FigAxes, self).save_data2(data)
        return data

    def load_data(self, fid=None):
        #        print "loading  fig_axes data"
        val = pickle.load(fid)
        for key in val:
            self.setp(key, val[key])
        super(FigAxes, self).load_data(fid)

    def load_data2(self, data):
        if 'FigAxes' in data:
            val = data['FigAxes'][1]

            # this if statement is bug fix recovery
            # normally this if is not executed
            if 'title_labelinfo' in val:
                if isinstance(val['title_labelinfo'], str):
                    dprint1(
                        'stored title_lableinfo is string!!!!. Executing recovery section')
                    # print self.getp('title_labelinfo')
                    self.getp('title_labelinfo')[0] = val['title_labelinfo']
                    del val['title_labelinfo']
                    # print self.getp('title_labelinfo')
                    if 'title_labelinfo' in data['FigObj'][2]:
                        del data['FigObj'][2]['title_labelinfo']
            for key in val:
                self.setp(key, val[key])
            if len(data['FigAxes']) > 2:
                param = data['FigAxes'][2]
                for key in param:
                    if hasattr(self, key):
                        setattr(self, key, param[key])
            if len(data['FigAxes']) > 3:
                axisdata = data['FigAxes'][3]
                self._xaxis = []
                for x in axisdata["xaxis"]:
                    ax = AxisXParam('x').import_data(self, x)
                    self._xaxis.append(ax)
                self._yaxis = []
                for x in axisdata["yaxis"]:
                    ax = AxisYParam('y').import_data(self, x)
                    self._yaxis.append(ax)
                self._zaxis = []
                for x in axisdata["zaxis"]:
                    ax = AxisZParam('z').import_data(self, x)
                    self._zaxis.append(ax)
                self._caxis = []
                for x in axisdata["caxis"]:
                    ax = AxisCParam('c').import_data(self, x)
                    self._caxis.append(ax)
        if not isinstance(self.getp('grid'), list):
            self.setp('grid', [False]*3)
        super(FigAxes, self).load_data2(data)

        from ifigure.ifigure_config import isMPL2
        if isMPL2:
            lp = self.getp("loaded_property")
            for x in lp:
                if 'axis_bgcolor' in x:
                    x['facecolor'] = x['axis_bgcolor']
                    del x['axis_bgcolor']

    def cla(self):
        from ifigure.mto.fig_text import FigText
        for name, child in self.get_children():
            if isinstance(child, FigAxes):
                continue
            if isinstance(child, FigText):
                continue
            child.destroy()

    def canvas_menu(self):
        return []

    def twinx(self, new_axesparam=True):
        if self.get_3d():
            return

        from ifigure.matplotlib_mod.axes_mod import AxesMod
        from ifigure.matplotlib_mod.axes3d_mod import Axes3DMod

        self._artists.append(AxesMod(self._artists[0].twinx()))
        if new_axesparam:
            p = self.add_axis_param(dir='y')
            if self._yaxis[0].tick_position == 'left':
                p.tick_position = 'right'
            else:
                p.tick_position = 'left'
            for p in self._yaxis:
                p.tick_both = False
#           self._axis_name_table[p.name] = len(self._artists)-1
        self._xaxis[0].add_ax_idx(len(self._artists)-1)
        self._yaxis[-1].add_ax_idx(len(self._artists)-1)
        self._artists[-1].figobj = self
        self._artists[-1].figobj_hl = []
        self._artists[-1].isTwin = True
        self._artists[-1].set_frame_on(False)
        self._artists[-1].set_zorder(self.getp('zorder'))
        self._hastwin[0] = True

        for a, p in zip(self._artists, self._yaxis):
            v = (p.name, self.get_ax_pos(a, p.name))
            self.set_ax_pos(v, a)
        return len(self._artists)-1

    def del_twinx(self):
        container = self.get_container()
        for p in self._xaxis + self._yaxis + self._zaxis + self._caxis:
            p.rm_ax_idx(len(self._artists)-1)
        container.delaxes(self._artists[-1])
        super(FigAxes, self).del_artist(artist=[self._artists[-1]])
        self._hastwin[0] = False
        self._yaxis = self._yaxis[:-1]
        for a, p in zip(self._artists, self._yaxis):
            v = (p.name, self.get_ax_pos(a, p.name))
            self.set_ax_pos(v, a)

    def twiny(self, new_axesparam=True):
        if self.get_3d():
            return

        from ifigure.matplotlib_mod.axes_mod import AxesMod
        from ifigure.matplotlib_mod.axes3d_mod import Axes3DMod

        self._artists.append(AxesMod(self._artists[0].twiny()))
        self._artists[-1].__class__ = self._artists[0].__class__
        if new_axesparam:
            p = self.add_axis_param(dir='x')
            if self._xaxis[0].tick_position == 'bottom':
                p.tick_position = 'top'
            else:
                p.tick_position = 'bottom'
            for p in self._xaxis:
                p.tick_both = False
#           self._axis_name_table[p.name] = len(self._artists)-1
        self._yaxis[0].add_ax_idx(len(self._artists)-1)
        self._xaxis[-1].add_ax_idx(len(self._artists)-1)
        self._artists[-1].figobj = self
        self._artists[-1].figobj_hl = []
        self._artists[-1].isTwin = True
        self._artists[-1].set_frame_on(False)
        self._artists[-1].set_zorder(self.getp('zorder'))
        self._hastwin[1] = True
        for a, p in zip(self._artists, self._xaxis):
            v = (p.name, self.get_ax_pos(a, p.name))
            self.set_ax_pos(v, a)
        return len(self._artists)-1

    def del_twiny(self):
        container = self.get_container()
        for p in self._xaxis + self._yaxis + self._zaxis + self._caxis:
            p.rm_ax_idx(len(self._artists)-1)
        container.delaxes(self._artists[-1])
        super(FigAxes, self).del_artist(artist=[self._artists[-1]])
        self._hastwin[1] = False
        self._xaxis = self._xaxis[:-1]
        for a, p in zip(self._artists, self._xaxis):
            v = (p.name, self.get_ax_pos(a, p.name))
            self.set_ax_pos(v, a)

    def set_twinx(self, value, a):
        if value[0]:
            i = self.twinx()
            for p, v in zip(self._yaxis, value[1]):
                p.import_data(self, v)
            self._yaxis[-1].set_artist_rangeparam(self._artists[i])
            for p in self._xaxis:
                for k in p._ax_idx:
                    p.set_artist_rangeparam(self._artists[k])
            self.apply_nomargin_mode()
        else:
            for p, v in zip(self._yaxis, value[1]):
                p.import_data(self, v)
            self.del_twinx()
            for p in self._xaxis:
                for k in p._ax_idx:
                    p.set_artist_rangeparam(self._artists[k])

    def set_twiny(self, value, a):
        if value[0]:
            i = self.twiny()
            for p, v in zip(self._xaxis, value[1]):
                p.import_data(self, v)
            self._xaxis[-1].set_artist_rangeparam(self._artists[i])
            for p in self._yaxis:
                for k in p._ax_idx:
                    p.set_artist_rangeparam(self._artists[k])
            self.apply_nomargin_mode()
        else:
            for p, v in zip(self._xaxis, value[1]):
                p.import_data(self, v)
            self.del_twiny()
            for p in self._yaxis:
                for k in p._ax_idx:
                    p.set_artist_rangeparam(self._artists[k])

    def get_twinx(self, a):
        value = [p.make_save_data(self) for p in self._yaxis]
        if self._hastwin[0]:
            return (True, value)
        return (False, value)

    def get_twiny(self, a):
        value = [p.make_save_data(self) for p in self._xaxis]
        if self._hastwin[1]:
            return (True, value)
        return (False, value)

    def hastwin(self, dir='both'):
        if dir.upper() == 'X':
            return self._hastwin[0]
        if dir.upper() == 'Y':
            return self._hastwin[0]
        if dir.upper() == 'BOTH':
            return [a for a in self._hastwin]

    def onAddTwinX(self, evt):
        #cvalue = self.get_twinx(None)
        self.twinx()
        value = [p.make_save_data(self) for p in self._yaxis]
        value.append(self._yaxis[-1].make_save_data(self))
        value[-1]["_ax_idx"] = []
        value = (True, value)
        #self.set_twinx(cvalue, None)
        self.del_twinx()
        h = UndoRedoFigobjMethod(self._artists[0], 'twinx', value)
        canvas = evt.GetEventObject()
        window = canvas.GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry([h])
        #ifigure.events.SendSelectionEvent(self.get_figpage(), canvas, canvas.selection)

    def onRemoveTwinX(self, evt):
        '''
        1) reorient clients to the remaining axes        
        2) remove ax artist
        '''
        value = [p.make_save_data(self) for p in self._yaxis]
        value = (False, value[:-1])
        h = []
        for child in self.walk_tree():
            if child.get_container_idx() == 1:
                h.append(UndoRedoFigobjMethod(child._artists[0],
                                              'container_idx', 0))
        h.append(UndoRedoFigobjMethod(self._artists[0], 'twinx', value))
        canvas = evt.GetEventObject()
        window = canvas.GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry(h)

    def onAddTwinY(self, evt):
        self.twiny()
        value = [p.make_save_data(self) for p in self._xaxis]
        value.append(self._xaxis[-1].make_save_data(self))
        value[-1]["_ax_idx"] = []
        value = (True, value)
        self.del_twiny()
        h = UndoRedoFigobjMethod(self._artists[0], 'twiny', value)
        canvas = evt.GetEventObject()
        window = canvas.GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry([h])

    def onRemoveTwinY(self, evt):
        value = [p.make_save_data(self) for p in self._xaxis]
        value = (False, value[:-1])
        h = []
        for child in self.walk_tree():
            if child.get_container_idx() == 1:
                h.append(UndoRedoFigobjMethod(child._artists[0],
                                              'container_idx', 0))

        h.append(UndoRedoFigobjMethod(self._artists[0], 'twiny', value))
        canvas = evt.GetEventObject()
        window = canvas.GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry(h)

    def set_caxis_value(self, value, a):
        while len(value) < len(self._caxis):
            self._caxis = self._caxis[:-1]
        while len(value) > len(self._caxis):
            ca_param = self.add_axis_param(dir='c')
        for p, v in zip(self._caxis, value):
            p.import_data(self, v)

    def get_caxis_value(self, a):
        value = [p.make_save_data(self) for p in self._caxis]
        return value

    def onAddC(self, evt):
        ca_param = self.add_axis_param(dir='c')
        value = [p.make_save_data(self) for p in self._caxis]
        self._caxis = self._caxis[:-1]
        h = []
        h.append(UndoRedoFigobjMethod(self._artists[0], 'caxis_value',
                                      value))
        canvas = evt.GetEventObject()
        window = canvas.GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry(h)

    def onRemoveC(self, evt):
        value = [p.make_save_data(self) for p in self._caxis[:-1]]
        h = []
        h.append(UndoRedoFigobjMethod(self._artists[0], 'caxis_value',
                                      value))
        canvas = evt.GetEventObject()
        window = canvas.GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry(h)

    def canvas_axes_menu(self):
        m = []
        if not self.get_3d():
            if (not self._hastwin[0] and not self._hastwin[1]):
                m.append(('Add X2 axis', self.onAddTwinY, None))
                m.append(('Add Y2 axis', self.onAddTwinX, None))
            if self._hastwin[0]:
                m.append(('Remove Y2 axis', self.onRemoveTwinX, None))
            if self._hastwin[1]:
                m.append(('Remove X2 axis', self.onRemoveTwinY, None))
        if len(self._caxis) == 0:
            txt = 'Add C axis'
        else:
            txt = 'Add C'+str(len(self._caxis)+1)+' axis'
        m.append((txt, self.onAddC, None))

        if len(self._caxis) > 0 and len(self._caxis[-1]._member) == 0:
            if len(self._caxis) == 1:
                txt = 'Remove C axis'
            else:
                txt = 'Remove C'+str(len(self._caxis))+' axis'
            m.append((txt, self.onRemoveC, None))
        if not isinstance(self, FigInsetAxes):
            m.append(('Add Inset Axes', self.onAddInsetAxesCanvas, None))
        if self.get_3d():
            m.append(('Use 2D Axes', self.onToggle3D, None))
        else:
            m.append(('Use 3D Axes', self.onToggle3D, None))

        return m

    def onToggle3D(self, evt):
        if self.get_3d():
            if sum([len(x._member) for x in self._zaxis]) != 0:
                from ifigure.widgets.dialog import message
                message(self, 'panel has 3D plot', 'error')
                return
            self.set_3d(False)
        else:
            self.set_3d(True)
        evt.GetEventObject().set_axes_selection(self._artists[0])
        ifigure.events.SendCanvasDrawRequest(evt.GetEventObject(),)
        ifigure.events.SendSelectionEvent(self, evt.GetEventObject(),
                                          evt.GetEventObject().selection)
#           figobj.reset_artist()

    def set_3d(self, value):
        if self._3D == value:
            return
        self._3D = value
        if self._3D and len(self._zaxis) == 0:
            self.add_axis_param(dir='z')
        if self._3D:  # going to 3d
            self.get_axis_bgfacecolor()
            self.get_axis_bgedgecolor()
            self.set_axis_bgalpha(0.0, None)
        if not self._3D:
            self._zaxis = []
        self.del_artist(delall=True)
        self.realize()
        if not self._3D:
            self.set_axis_bgedgecolor(self.getp('axis_bgedgecolor'))
            self.set_axis_bgfacecolor(self.getp('axis_bgfacecolor'))
            self.set_axis_bgalpha(1.0, None)
        self.set_bmp_update(False)

    def get_3d(self):
        return self._3D

    def set_axis3d_bgcolor(self, value, artist):
        self.set_3d_pane_colorname(value)

    def get_axis3d_bgcolor(self, artist):
        return self.get_3d_pane_colorname()

    def set_axis3d_bgalpha(self, value, artist):
        if value is None:
            value = 1.0
        self.set_3d_pane_alpha([float(value)]*3)

    def get_axis3d_bgalpha(self, artist):
        return self.get_3d_pane_coloralpha()[0]

    def get_gl_lighting(self, a):
        loc = a._lighting['light_direction']
        phi = np.arctan2(loc[1], loc[0])
        theta = np.arctan2(np.sqrt(loc[1]**2 + loc[0]**2), loc[2])

        v = [None,
             a._lighting['ambient'],
             a._lighting['light'],
             a._lighting['specular'],
             None,
             phi*180/np.pi,
             theta*180/np.pi,
             a._lighting['shadowmap'],
             a._use_frustum]
        return v

    def set_gl_lighting(self, v, a):
        a._lighting['ambient'] = v[1]
        a._lighting['light'] = v[2]
        a._lighting['specular'] = v[3]
        a._lighting['shadowmap'] = v[7]
        phi, th = v[5]*np.pi/180., v[6]*np.pi/180.
        loc = np.array([np.sin(th)*np.cos(phi),
                        np.sin(th)*np.sin(phi),
                        np.cos(th)])
        a._lighting['light_direction'] = loc
        a._use_frustum = v[8]

    def set_axes3d_viewparam(self, value, a, no_proj=False):
        if len(value) == 2:
            elev, azim = value
        else:
            elev, azim, upvec = value
            a._upvec = upvec
        a.elev = elev
        a.azim = azim
        if not no_proj:
            a.get_proj()
        self.set_bmp_update(False)

    def get_axes3d_viewparam(self, ax):
        return ax.elev, ax.azim, ax._upvec

    def set_axis_bgalpha(self, value, artist=None):
        for a in self._artists:
            a.patch.set_alpha(0.0)
        self._artists[0].patch.set_alpha(value)
        self.setp('axis_bgalpha', value)
        # self.set_3d_pane_alpha([float(value)]*3)

    def get_axis_bgalpha(self, artist=None):
        return self._artists[0].patch.get_alpha()

    def set_axis_bgedgealpha(self, value, artist=None):
        for a in self._artists:
            a.patch.set_linewidth(0.0)
        self._artists[0].patch.set_linewidth(value)
        self.setp('axis_bgalpha', value)
        # self.set_3d_pane_alpha([float(value)]*3)

    def get_axis_bglinewidth(self, artist):
        return self._artists[0].patch.get_linewidth()

    def set_axis_bglinewidth(self, value, artist=None):
        for a in self._artists:
            a.patch.set_linewidth(value)
        self.setp('axis_bglinewidth', value)
        # self.set_3d_pane_alpha([float(value)]*3)

    def get_axis_bglinestyle(self, artist):
        return self._artists[0].patch.get_linestyle()

    def set_axis_bglinestyle(self, value, artist=None):
        for a in self._artists:
            a.patch.set_linestyle(value)
        self.setp('axis_bglinestyle', value)
        # self.set_3d_pane_alpha([float(value)]*3)

    def get_axis_bgedgecolor(self, artist=None):
        v1 = self._artists[0].patch.get_edgecolor()
        if v1 != self.getp('axis_bgedgecolor'):
            self.setp('axis_bgedgecolor', v1)
        return v1

    def set_axis_bgedgecolor(self, value, artist=None):
        for a in self._artists:
            a.patch.set_edgecolor((0, 0, 0, 0))
        self._artists[0].patch.set_edgecolor(value)
        self.setp('axis_bgedgecolor', value)
        # self.set_3d_pane_alpha([float(value)]*3)

    def get_axis_bgfacecolor(self, artist=None):
        v1 = self._artists[0].patch.get_facecolor()
        if v1 != self.getp('axis_bgfacecolor'):
            self.setp('axis_bgfacecolor', v1)
        return v1

    def set_axis_bgfacecolor(self, value, artist=None):
        alpha = self._artists[0].patch.get_alpha()
        for a in self._artists:
            #            a.set_axis_bgcolor((0,0,0,0))
            a.patch.set_facecolor((0, 0, 0, 0))
            a.patch.set_alpha(0)
        value = list(value)
        if not any(value):
            alpha = 0.0
        if alpha is None:
            alpha = 1.0
        value[3] = alpha
        self._artists[0].patch.set_facecolor(value)
        self._artists[0].patch.set_alpha(value[3])
#        self._artists[0].set_axis_bgcolor(value)
        self.setp('axis_bgfacecolor', value)
        # self.set_3d_pane_alpha([float(value)]*3)

    def set_3d_pane_colorname(self, value):
        a = self.get_3d_pane_coloralpha()
        self._3d_pane_color = self._3d_pane_var(value, a)
        self._set_3d_pane_color(self.get_3d_pane_colorcode())

    def set_3d_pane_alpha(self, a):
        value = self.get_3d_pane_colorname()
        self._3d_pane_color = self._3d_pane_var(value, a)
        self._set_3d_pane_color(self.get_3d_pane_colorcode())

    def get_3d_pane_colorname(self):
        return list(zip(*self._3d_pane_color))[0]

    def get_3d_pane_coloralpha(self):
        return list(zip(*self._3d_pane_color))[1]

    def get_3d_pane_colorcode(self):
        return list(zip(*self._3d_pane_color))[2]

    def _set_3d_pane_color(self, code):
        from ifigure.matplotlib_mod.axes_mod import AxesMod
        from ifigure.matplotlib_mod.axes3d_mod import Axes3DMod

        if len(self._artists) == 0:
            return
        if not isinstance(self._artists[0], Axes3DMod):
            return
        a = self._artists[0]
        a.xaxis.set_pane_color(code[0])
        a.yaxis.set_pane_color(code[1])
        a.zaxis.set_pane_color(code[2])

    def _3d_pane_var(self, value, a):
        from matplotlib.colors import ColorConverter as cc
        return [(value[0], a[0], cc().to_rgba(value[0], alpha=a[0])),
                (value[1], a[1], cc().to_rgba(value[1], alpha=a[1])),
                (value[2], a[2], cc().to_rgba(value[2], alpha=a[2]))]

    def set_ax_pos(self, value, a):
        p = self.get_axis_param(value[0])
        p.tick_position = value[1][0]
        p.tick_both = value[1][1]
        p.box = value[1][2]

        #        a = self.get_axes_artist_by_name(value[0])
        for a in self.get_axes_artist_by_name(value[0]):
            p.set_box(a)
            if value[0].startswith('x'):
                p.set_artist_tickposition(a)
                if p.tick_both:
                    a.xaxis.set_ticks_position('both')
            elif value[0].startswith('y'):
                p.set_artist_tickposition(a)
                if p.tick_both:
                    a.yaxis.set_ticks_position('both')

    def get_ax_pos(self, a, name):
        p = self.get_axis_param(name)
        value = (p.tick_position,
                 p.tick_both,
                 p.box,
                 p.get_artist_tickposition(a))
        return value

    def set_cmap3(self, value, a):
        p = self.get_axis_param(value[0])
        p.set_cmap(value[1])
        for a in self._artists:
            p.set_artist_rangeparam(a)
        self.set_bmp_update(False)
        p.update_cb()

    def get_cmap3(self, value, name):
        p = self.get_axis_param(name)
        value = p.get_cmap()
        return value

    def reset_color_cycle(self):
        from ifigure.matplotlib_mod.mpl_utils import reset_color_cycle
        for a in self._artists:
            reset_color_cycle(a)

    def get_auto_status_str(self):
        area = self.getp('area')
        f = '{:.2f}'
        txt = [f.format(a) for a in area]
        return 'L:' + txt[0] + ' B:'+txt[1]+' W:' + txt[2] + ' H:' + txt[3]

    def convert_to_tex_style_text(self, mode=True):
        from ifigure.utils.cbook import tex_escape_equation
        if len(self._artists) == 0:
            return
        labels = []
        for a in self._artists:
            labels.extend([a.title, a.xaxis.label, a.yaxis.label])
            if self._3D:
                labels.append(a.zaxis.label)
        if mode:
            s_bk = [l.get_text() for l in labels]
            self.setp('s_bk', s_bk)
            for s, l in zip(s_bk, labels):
                l.set_text(tex_escape_equation(str(s)))
        else:
            s_bk = self.getp('s_bk')
            for s, l in zip(s_bk, labels):
                l.set_text(s)


class FigInsetAxes(FigAxes):
    def __init__(self, *args, **kywds):
        #        self.child=[]
        if "draggable" in kywds:
            self.setvar("draggable", kywds["draggable"])
            del kywds["draggable"]
        else:
            self.setvar("draggable", True)

        super(FigInsetAxes, self).__init__(*args, **kywds)

        self.setp("inset_w", 0.45)
        self.setp("inset_h", 0.45)
        self.setp("inset_transform", 'axes')
        self.setp("inset_anchor", (0.05, 0.5))
        self._floating = True

    @property
    def ispickable_a(self):
        return True

    @classmethod
    def isFigInsetAxes(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'insetaxes'

    @classmethod  # define _attr values to be saved
    def attr_in_file(self):
        return super(FigInsetAxes, self).attr_in_file()+["inset_w", "inset_h", "inset_transform", "inset_anchor"]

    @classmethod
    def property_in_palette(self):
        return ["inset_switchtrans", "inset_size"]

    def tree_viewer_menu(self):
     # return MenuString, Handler, MenuImage
     #  m=[('Add Plot',     self.onAddPlot, None),
     #     ('Add Contour',  self.onAddContour, None),
     #     ('Add Image',    self.onAddImage, None),
     #     ('---',        None, None),
     #     ('Realize',    self.onRealize, None)]
        return super(FigAxes, self).tree_viewer_menu()

    def isDraggable(self):
        return self._var["draggable"]

    def set_parent(self, parent):
        super(FigInsetAxes, self).set_parent(parent)
        if parent is None:
            return
        if hasattr(parent, 'add_update_client'):
            parent.add_update_client(self)

    def get_container(self):
        #        f_axes = self._parent
        #        return f_axes._parent._artists[0]
        figpage = self.get_figpage()
        return figpage._artists[0]

    def do_update_artist(self):
        #        print 'do-update inset_axes'
        rect, use_def, margin = self.calc_rect()
        for a in self._artists:
            a.set_position(rect)
        self.set_client_update_artist_request()
        if self.has_highlight(self._artists[0]):
            self.highlight_artist(False, artist=[self._artists[0]])
            self.highlight_artist(True, artist=[self._artists[0]])
        self.set_bmp_update(False)

    def get_inset_transform(self, name):
        fig_axes = self._parent
        fig_page = self.get_figpage()

        return self.coordsname2transform(fig_page, name,
                                         fig_axes=fig_axes)

    def convert_transform(self, name1, name2, rect):

        if name1 is not None:
            t1 = self.get_inset_transform(name1)
        else:
            t1 = None
        if name2 is not None:
            t2 = self.get_inset_transform(name2)
        else:
            t2 = None
#        print self._parent._artists[0].get_window_extent().get_points()
        v = self.convert_transform_rect(t1, t2, rect)
        return v

    def calc_rect(self, ignore_pagemargin=False):
        #ignore_pagemargin is not used
        ax_main = self._parent
        rect, use_def, margin = ax_main.calc_rect()
        w = self.getp("inset_w")
        h = self.getp("inset_h")
        anchor = self.getp("inset_anchor")

        rect = [anchor[0], anchor[1], w, h]
        rect = self.convert_transform(
            self.getp("inset_transform"), 'figure', rect)
#        print 'rect', rect
        self._rect = rect
        return rect, use_def, margin

    def set_area(self, area):
        print("inset_axes set_area is called. this might be due to a potential bug.")
        pass

    def highlight_artist(self, val, artist=None):
        if artist is None:
            alist = self._artists
        else:
            alist = artist
        if val == True:

            for a in alist:
                box = a.get_window_extent().get_points()
                x = [box[0][0], box[0][0], box[1][0], box[1][0], box[0][0]]
                y = [box[0][1], box[1][1], box[1][1], box[0][1], box[0][1]]
                hl = Line2D(x, y, marker='s',
                            color='k', linestyle='None',
                            markerfacecolor='None',
                            markeredgewidth=0.5,
                            figure=a.figure)
                a.figure.lines.extend([hl])
                a.figobj_hl.append(hl)
        else:
            for a in alist:
                for hl in a.figobj_hl:
                    a.figure.lines.remove(hl)
                a.figobj_hl = []
#
#  Hit Test
#

    def get_artist_extent(self, a):
        box = a.get_window_extent().get_points()
        return [box[0][0], box[1][0], box[0][1], box[1][1]]

    def picker(self, a, evt):
        return self.picker_a(a, evt)

    def dragstart_a(self, a, evt, shift=None):
        redraw = super(FigInsetAxes, self).dragstart_a(a, evt)
        if self._drag_hl is not None:
            self._drag_hl.set_linestyle('-')
            self._drag_hl.set_color('r')
            self._drag_hl.set_alpha(0.5)

    def dragdone_a(self, a, evt, shift=None, scale=None):
        redraw = super(FigInsetAxes, self).dragdone_a(a, evt)

        anchor = [min([self._drag_rec[0], self._drag_rec[1]]),
                  min([self._drag_rec[2], self._drag_rec[3]])]
        w = abs(self._drag_rec[0] - self._drag_rec[1])
        h = abs(self._drag_rec[2] - self._drag_rec[3])

        rect = [anchor[0], anchor[1], w, h]
        rect = self.convert_transform(None, self.getp("inset_transform"), rect)

        h = [UndoRedoFigobjMethod(a, 'inset_rect', rect)
             for a in self._artists]
        canvas = evt.guiEvent_memory.GetEventObject()
        window = canvas.GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry(h,
                                                       menu_name='move')
        return True, None

    def scale_artist(self, scale, action=True):
        rect = self.get_inset_rect()
        rect = self.convert_transform(self.getp("inset_transform"), None, rect)
        # this rect is in device coords.
        area = scale_rect([rect[0], rect[0]+rect[2],
                           rect[1], rect[1]+rect[3]], scale)
        rect2 = [min(area[:2]), min(area[2:]), abs(area[1] - area[0]),
                 abs(area[3] - area[2])]
        rect = self.convert_transform(None, self.getp("inset_transform"),
                                      rect2)
        action1 = []
        for a in self._artists:
            action1.append(UndoRedoFigobjMethod(a, 'inset_rect', rect))
        return action1

    def set_inset_rect(self, value, a):
        self.setp('inset_anchor', value[0:2])
        self.setp('inset_w', value[2])
        self.setp('inset_h', value[3])
        self.do_update_artist()

    def get_inset_rect(self, a=None):
        x, y = self.getp('inset_anchor')
        w = self.getp('inset_w')
        h = self.getp('inset_h')
        return [x, y, w, h]

    def set_insetsize(self, value, a):
        v = [float(x) for x in value.split(',')]
        self.set_inset_rect(v, a)

    def get_insetsize(self, a=None):
        return ', '.join([str(x) for x in self.get_inset_rect()])

    def get_switchtrans(self, a):
        return self.getp('inset_transform')

    def set_switchtrans(self, value, a):
        v = str(value)

        w = self.getp("inset_w")
        h = self.getp("inset_h")
        anchor = self.getp("inset_anchor")
        rect = [anchor[0], anchor[1], w, h]

#        print rect
        rect = self.convert_transform(self.getp("inset_transform"), v, rect)
#        print rect
        self.setp("inset_w",  rect[2])
        self.setp("inset_h",  rect[3])
        self.setp("inset_anchor",  rect[0:2])

        return self.setp('inset_transform', v)


class FigColorBar(FigInsetAxes, AdjustableRangeHolderCbar):
    def __init__(self, *args, **kywds):
        super(FigColorBar, self).__init__(*args, **kywds)
        self._generic_axes = False
        self._cbar_image = None
        self._cbclient = []
        self._caxis_param = None
        self.setp("cdir", "v")

        self.setp("inset_w", 0.05)
        self.setp("inset_h", 0.8)
        self.setp("inset_anchor", (0.9, 0.1))
    # ColorBar does not support cursor...

    def generate_cursor(self, evt, idx):
        # called during mouse click in cursor mode
        return []

    @classmethod
    def property_in_palette(self):
        return ["inset_size"]

    def update_cursor(self, evt, idx):
        # called during mousedrag in cursor mode
        pass

    def valid_cursor(self):
        # called just before drawing cursor
        return []

    def canvas_axes_menu(self):
        return []

    def set_caxis_param(self, param):
        self._caxis_param = weakref.ref(param)

    def property_in_palette_axes(self):
        names = [x.name for x in (self._xaxis + self._yaxis)]
        tab = ["common"]+[self._caxis_param().name] + names
        item = [["title", "frame", "axis", ], ]
        item.append(["axcrangeparam13", "axformat"])
        for i in self._xaxis:
            item.append(["axlabel",
                         "axlotsize",
                         "axtlcolor", "axticks",
                         "axxpos"])
        for i in self._yaxis:
            item.append(["axlabel",
                         "axlotsize",
                         "axtlcolor", "axticks", "axypos"])
        return tab, item

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'colorbar.png')
        return [idx1]

    @classmethod
    def get_namebase(self):
        return 'colorbar'

    def destroy(self, clean_owndir=True):
        if self._cbar_image is not None:
            self._cbar_image.remove()
            self._cbar_image = None
        self._caxis_param = None
        super(FigColorBar, self).destroy(clean_owndir=clean_owndir)

    def generate_artist(self, *args, **kywds):
        super(FigColorBar, self).generate_artist(*args, **kywds)
        if self._caxis_param is None:
            return
        self.update_cbar_image()

    def del_artist(self, *args, **kywds):
        if self._cbar_image is not None:
            self._cbar_image.remove()
            self._cbar_image = None
        self.remove_2nd_artist()
        super(FigColorBar, self).del_artist(*args, **kywds)

    def remove_2nd_artist(self):
        if len(self._artists) == 2:
            artist = self._artists[1]
            artist.figobj = None
            artist.figobj_hl = []
            artist.remove()
            self._artists = [self._artists[0]]

    def update_cbar_image(self):
        if self._caxis_param is None:
            return
        dprint2('update_cbar_image')
        array = np.vstack((np.arange(256), np.arange(256)))
        self.set_cbarscale()
        cmesh = self.cmesh()
        if self._cbar_image is not None:
            self._cbar_image.remove()
            self._cbar_image = None

        a = self._artists[0]

        scale = self._caxis_param().scale
        zp = np.vstack((cmesh, cmesh))

        if self.getp('cdir') == 'h':
            self._artists[0].set_ylim((0, 1))
            y = (0, 1)
            a.set_xlim((cmesh[0], cmesh[-1]))
            a.set_autoscalex_on(False)
            a.set_autoscaley_on(False)
            self.set_axtcolor(['y', [0, 0, 0, 0]], a)
            self.set_axlcolor(['y', ([0, 0, 0, 0], [0, 0, 0, 0])], a)
            lc = self._caxis_param().lcolor
            tc = self._caxis_param().tcolor
            self.set_axtcolor(['x', tc], a)
            self.set_axlcolor(['x', (lc, lc)], a)
            extent = [cmesh[0], cmesh[-1], min(y), max(y)]

        elif self.getp('cdir') == 'v':
            x = self._artists[0].set_xlim((0, 1))
            x = (0, 1)
            a.set_ylim((cmesh[0], cmesh[-1]))
            a.set_autoscalex_on(False)
            a.set_autoscaley_on(False)
            zp = np.transpose(zp)
            self.set_axtcolor(['x', [0, 0, 0, 0]], a)
            self.set_axlcolor(['x', ([0, 0, 0, 0], [0, 0, 0, 0])], a)
            lc = self._caxis_param().lcolor
            tc = self._caxis_param().tcolor
            self.set_axtcolor(['y', tc], a)
            self.set_axlcolor(['y', (lc, lc)], a)
            extent = [min(x), max(x), cmesh[0], cmesh[-1]]

        if len(self._artists) > 1:
            self._artists[1].remove()
            self._artists = [self._artists[0], ]
        self.generate_twin_artist()

        self._cbar_image = self._artists[1].imshow(zp,
                                                   extent=extent,
                                                   interpolation='bicubic',
                                                   resample=True,
                                                   aspect='auto',
                                                   origin='lower')
        self._cbar_image.nozsort = True
        self.call_set_crangeparam_to_artist(self._caxis_param())

        if self._caxis_param().cmap is not None:
            self._cbar_image.set_cmap(self._caxis_param().cmap)

    def generate_twin_artist(self):
        if self.getp('cdir') == 'v':
            artist = self._artists[0].twinx()
            artist.yaxis.set_major_locator(mticker.NullLocator())
            artist.yaxis.set_major_formatter(mticker.NullFormatter())
        else:
            artist = self._artists[0].twiny()
            artist.xaxis.set_major_locator(mticker.NullLocator())
            artist.xaxis.set_major_formatter(mticker.NullFormatter())
        artist.figobj = self
        artist.figobj_hl = []
        artist.set_zorder(self.getp('zorder'))
        self._artists.append(artist)

    def set_update_cbar_image(self, value, a):
        self.update_cbar_image()

    def get_update_cbar_image(self, a):
        pass

    def set_cbarscale(self):
        if self._caxis_param is None:
            return
        scale = self._caxis_param().scale
        a = self._artists[0]
        a.set_yscale('linear')
        a.set_xscale('linear')
        if scale == 'log':
            if self.getp('cdir') == 'v':
                a.set_yscale('log')
            else:
                a.set_xscale('log')
        elif scale == 'symlog':
            th = self._caxis_param().symloglin
            range = self._caxis_param().range

            if self.getp('cdir') == 'v':
                a.set_ylim(range)
                if isMPL33:
                    a.set_yscale('symlog', linthresh=th)
                else:
                    a.set_yscale('symlog', linthreshy=th)
            else:
                a.set_xlim(range)
                if isMPL33:
                    a.set_xscale('symlog', linthresh=th)
                else:
                    a.set_xscale('symlog', linthreshx=th)
        if self.getp('cdir') == 'v':
            p = self.get_axis_param('y')
        else:
            p = self.get_axis_param('x')
        p.scale = scale

    def cmesh(self):
        param = self._caxis_param()
        cmin, cmax = param.range
        scale = param.scale

        if scale == 'linear':
            cmesh = np.linspace(cmin, cmax, 256)
        elif scale == 'symlog':
            cmesh = np.linspace(cmin, cmax, 256)
        else:
            #
            cmin = cmin if cmin > 0.0 else cmax*1e-16
            cmesh = np.linspace(cmin, cmax, 256)

        return cmesh

    def canvas_menu(self):
        return [("Colormap...",  self.onColormap, None),
                ("Clipping...",  self.onClip, None),
                ("Auto Scale C",  self.onAutoC, None),
                ("Switch Direction",  self.onDir, None)]

    def onDir(self, evt=None):
        if self.getp('cdir') == 'v':
            value = 'h'
        else:
            value = 'v'

        h = [UndoRedoFigobjMethod(self._artists[0], 'cbardir',
                                  value)]
        canvas = evt.GetEventObject()
        window = canvas.GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry(h,
                                                       menu_name='cbar direction')

    def set_cbardir(self, value, a):
        pvalue = self._caxis_param().get_rangeparam()
        lim = pvalue[2]
        if value == 'v':
            self.setp('cdir', value)
#           self._ycolor=(self.get_ytcolor(self._artists[0]),
#                         self.get_ylcolor(self._artists[0]))
            self._artists[0].set_ylim((0, 1))
            self._artists[0].set_xlim(lim)
        elif value == 'h':
            self.setp('cdir', value)
            self._artists[0].set_xlim((0, 1))
            self._artists[0].set_ylim(lim)
        self.remove_2nd_artist()
        self.update_cbar_image()
        ifigure.events.SendPVDrawRequest(self)

    def get_cbardir(self, a):
        return self.getp('cdir')

    def onAutoC(self, evt):
        canvas = evt.GetEventObject()
        canvas.set_axes_selection(self._parent._artists[0])
        self.hndl_dclick(canvas)

    def onClip(self, evt):
        if self._caxis_param is None:
            return

        ca = self._caxis_param()
        canvas = evt.GetEventObject()
        l = [["", "Clip Setting", 2],
             ["Under", ca.clip[0], 3006, ({"text": "Clip"}, {})],
             ["Over", ca.clip[1], 3006, ({"text": "Clip"}, {})]]
        tip = None
        value = DialogEditList(l, tip=tip, parent=canvas,
                               style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)
        if value[0]:
            artist = self._artists[0]
            actions = [UndoRedoFigobjMethod(artist,
                                            'clip', (value[1][1], value[1][2]))]
            window = canvas.GetTopLevelParent()
            GlobalHistory().get_history(window).make_entry(actions)

#            ca.clip = (value[1][1], value[1][2])
#            for a in self._parent._artists:
#                ca.set_artist_rangeparam(a)
#            self._parent.set_bmp_update(False)
#            canvas.draw()

    def set_clip(self, value, a0):
        ca = self._caxis_param()
        ca.clip = value
        for a in self._parent._artists:
            ca.set_artist_rangeparam(a)
        self._parent.set_bmp_update(False)

    def get_clip(self, a):
        ca = self._caxis_param()
        return ca.clip

    def set_axcrangeparam13(self, value, a0):
        return self._parent.set_axrangeparam13(value, a0)

    def get_axcrangeparam13(self, a, name):
        return self._parent.get_axrangeparam13(a, name)

    def onColormap(self, evt):
        ca = self._caxis_param()
        canvas = evt.GetEventObject()

        name = ca.cmap
        if name.endswith('_r'):
            s = {"reverse": True}
        else:
            s = {"reverse": False}
        l = [["", name, 11, s]]
        value = DialogEditList(l, tip=None, parent=canvas)
        if value[0]:
            artist = self._artists[0]
            actions = [UndoRedoFigobjMethod(artist,
                                            'cmap', value[1][0])]
            window = canvas.GetTopLevelParent()
            GlobalHistory().get_history(window).make_entry(actions)

    def set_cmap(self, value, a0):
        ca = self._caxis_param()
        ca.cmap = value
        self._cbar_image.set_cmap(ca.cmap)
        self.call_set_crangeparam_to_artist(ca)

        for a in self._parent._artists:
            ca.set_artist_rangeparam(a)
        self._parent.set_bmp_update(False)
        self.set_bmp_update(False)

    def get_cmap(self, a0):
        ca = self._caxis_param()
        return ca.cmap

    def call_set_crangeparam_to_artist(self, ca):
        scale = ca.scale
        ca.scale = 'linear'
        ca.set_crangeparam_to_artist(self._cbar_image, check=False)
        ca.scale = scale

    def handle_axes_change(self, data):
        #        print 'handle_axes_change'
        evtTD = data['td']
#        hist = self.get_root_parent().app.history
        dprint2('handle_axes_changed in cbar')
        if evtTD == self:
            return
        else:
            self.update_cbar_image()
            self.set_bmp_update(False)

    def tree_viewer_menu(self):
     # return MenuString, Handler, MenuImage
        m = []
        return m+super(FigAxes, self).tree_viewer_menu()

    def hndl_dclick(self, canvas):
        '''
        called when user double click the fig_axes
           1) make a dummy request to make axes
           3) make a new request similar to range zoom 
        '''
        canvas.set_axes_selection(self._parent._artists[0])
        x = self._caxis_param()
        value = x.get_rangeparam()
        value[1] = True
        request = [(x.name, value)]
        requests = {
            self._parent: self._parent.compute_new_range(request=request)}
        requests = canvas.expand_requests(requests)
        canvas.send_range_action(requests, 'color range')

    def save_data2(self, data):
        data = super(FigColorBar, self).save_data2(data)

        index = self._parent._caxis.index(self._caxis_param())
        cbdata = {"caxisparam_index": index}
        data['FigColorBar'] = (1, cbdata)

        return data

    def load_data2(self, data):
        if 'FigColorBar' in data:
            cbdata = data["FigColorBar"][1]
            self._loaded_cbdata = cbdata
        return super(FigColorBar, self).load_data2(data)

    def init_after_load(self, olist, nlist):
        if hasattr(self, '_loaded_cbdata'):
            cbdata = self._loaded_cbdata
            if "caxisparam_index" in cbdata:
                k = cbdata["caxisparam_index"]
                self.set_caxis_param(self._parent._caxis[k])
            del self._loaded_cbdata

    def _get_caxis(self):
        if self.getp("cdir") == 'v':
            axis = a.get_yaxis()
        else:
            axis = a.get_xaxis()
        return axis

    def set_adjustrange(self, *args):
        pass

    def get_adjustrange(self, *args):
        pass

    # format
    def set_axformat(self, param, a):
        ca = self._caxis_param()
        ca.format = param[1]
        cdir = self.getp('cdir')
        if cdir == 'v':
            name = 'y'
        else:
            name = 'x'
        p = self.get_axis_param(name)
        p.format = param[1]
        p.set_ticks(self.get_axes_artist_by_name(name)[0])
        # p.set_ticks(self.get_axes_artist_by_name(param[0])[0])
        self.set_bmp_update(False)

    def get_axformat(self, a, name):
        ca = self._caxis_param()
        return ca.format



from __future__ import print_function
#  Name   :py_solver
#
#          various classes for defining solvers
#
#  Author :
#         Syun'ichi Shiraiwa
#  E-mail :
#         shiraiwa@psfc.mit.edu
#
#
#  History : 2012 08 20 : extracted from py_code
#
# *******************************************
#     Copyright(c) 2012- PSFC, MIT
# *******************************************

import os
import wx
import time
import sys
import logging
import weakref
import shutil
import threading
import numpy as np
import ifigure.utils.pickle_wrapper as pickle
import ifigure
import ifigure.utils.cbook as cbook
import ifigure.widgets.dialog as dialog
from ifigure.mto.py_code import PyCode, PyModel, PyParam, PySol, AbsModuleContainer, AbsScriptContainer
from ifigure.mto.py_script import PyScript
from ifigure.mto.py_module import PyModule
import ifigure.events


from numpy import linspace


def onSettingList(solver):
    app = wx.GetApp().TopWindow

    model = ''
    if cbook.ProxyAlive(solver._model):
        model = solver._model.get_full_path()
    else:
        solver._model = None

    sol = ''
    if cbook.ProxyAlive(solver._sol):
        sol = solver._sol.get_full_path()
    else:
        solver._sol = None

    param = ''
    if cbook.ProxyAlive(solver._param):
        param = solver._param.get_full_path()
    else:
        solver._param = None

    list = [["", "Solver Setting", 2],
            ["Model", model, 0],
            ["Solution", sol, 0],
            ["Global", param, 0], ]
    return list


def setSetting(solver, value):
    app = solver.get_app()
    try:
        root = solver.get_root_parent()
        exec(root._name + ' = root')
        if len(value[1].split()) != 0:
            a = eval(value[1])
            if isinstance(a, PyModel):
                solver._model = weakref.proxy(a)
            else:
                dialog.message(app, 'Model should be PyModel', 'Error')
                solver._model = None
        if len(value[2].split()) != 0:
            b = eval(value[2])
            if isinstance(b, PySol):
                solver._sol = weakref.proxy(b)
            else:
                dialog.message(app, 'Sol should be PySol', 'Error')
                solver._sol = None
        if len(value[3].split()) != 0:
            b = eval(value[3])
            if isinstance(b, PyParam):
                solver._param = weakref.proxy(b)
            else:
                dialog.message(app, 'Parameter should be PyParam', 'Error')
                solver._param = None
    except Exception:
        logging.exception("Solver Setting Failed")


def init_sol(solver, sol=None, mode=0, tmpdir=None):
    '''
    mode = 0: standard mode
    mode = 1: mode 0 + backup tmpdir and return it to caller
    mode = 2: read data from tmpdir
    '''
    # cleaning folder
    if not cbook.ProxyAlive(solver._model):
        print("can not initialize solution folder. model is not spedified")
        solver._model = None
        return False

    if sol is None:
        if cbook.ProxyAlive(solver._sol):
            sol = solver._sol

    if sol is None:
        print("can not initialize solution folder. sol is not spedified")
        return False

    for name, child in sol.get_children():
        child.destroy()

    root = solver.get_root_parent()
    wdir = root.getvar('wdir')

    if mode != 2:
        fpath = os.path.join(wdir, '.tmp'+str(solver.id))
        if os.path.exists(fpath):
            os.remove(fpath)
        rm = solver._model.get_root_model()

        scripts = [obj for obj in rm.walk_tree() if isinstance(obj, PyScript)]
        for s in scripts:
            s._save_link = True
        flag, tmpdir = rm.save_subtree(fpath, maketar=False)
        for s in scripts:
            s._save_link = False
    else:
        shutil.copytree(tmpdir+'_bk', tmpdir)
    if mode == 1:
        shutil.copytree(tmpdir, tmpdir+'_bk')
    sol.load_subtree(tmpdir, usetar=False)
    sol.setvar('source_model', solver._model.get_root_model().get_full_path())
#        os.remove(fpath)

    if not sol.has_child('parameters'):
        sol.add_child('parameters', PyParam())
    if cbook.ProxyAlive(solver._param):
        #            psetting = root.setting.parameters
        param = solver._param
        vars = param.getvar_copy()
        sol.parameters.setvar(vars)
    if mode == 1:
        return True, tmpdir
    return True


def check_ref(solver):
    if not hasattr(solver._model, 'isTreeDict'):
        solver._model = None
    if not hasattr(solver._sol, 'isTreeDict'):
        solver._sol = None
    if not hasattr(solver._param, 'isTreeDict'):
        solver._param = None


class PySolver(PyCode, AbsModuleContainer, AbsScriptContainer):
    '''
     PySolver is a top node of solver tree 
    '''

    def __init__(self, parent=None, src=None):
        self._model = None
        self._sol = None
        self._param = None
        self._thread = None
        super(PySolver, self).__init__(parent=None, src=None)

    @classmethod
    def isPySolver(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'solver'

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'solver.png')
        return [idx1]

    def pv_kshortcut(self, key):
        '''
        define keyborad short cut on project viewer
        must return a method responding key
        '''
        if key == wx.WXK_F9:
            return self.onRun
        return None

    def get_pysol(self, parent):
        return [child for name, child in parent.get_children()
                if isinstance(child, PySol)]

    def onAddSolver(self, e=None):
        '''
        add generic solver object.
        this option is used when building
        solver tree from scratch.
        '''
        sc = BaseSolver()
        name = self.get_next_name(sc.get_namebase())
        self.add_child(name, sc)
        if e is not None:
            e.Skip()

    def onAddStdSolver(self, e=None):
        sc = PyStdSolver()
        name = self.get_next_name(sc.get_namebase())
        self.add_child(name, sc)
        sc.load_default_script()
        if e is not None:
            e.Skip()

    def onAddParametric(self, e=None):
        sc = PyParametric()
        name = self.get_next_name(sc.get_namebase())
        self.add_child(name, sc)
        sc.load_default_script()
        if e is not None:
            e.Skip()

    def onAddOptimizer(self, e=None):
        sc = PyOptimizer()
        name = self.get_next_name(sc.get_namebase())
        self.add_child(name, sc)
        sc.load_default_script()
        if e is not None:
            e.Skip()

    def tree_viewer_menu(self):
     # return MenuString, Handler, MenuImage
        return [  # ('Run   (F9)', self.onRun, None),
            ('+Add Solver', None, None),
            ('Standard Solver', self.onAddStdSolver, None),
            ('Parametric Solver', self.onAddParametric, None),
            ('Optimizer', self.onAddOptimizer, None),
            ('Custom Solver', self.onAddSolver, None),
            ('!',  None, None),
            ('---', None, None)] + \
            super(PyCode, self).tree_viewer_menu()

    def module_addon_list(self):
        base_mod = 'ifigure.add_on.solver.module'
        mod_path = cbook.GetModuleDir(base_mod)
        menu = self.build_module_menu(mod_path)
        menu.append(('New Module ...', self.onAddNewModule, None))
        return menu

    def script_addon_list(self):
        base_mod = 'ifigure.add_on.solver.script'
        mod_path = cbook.GetModuleDir(base_mod)
        menu = self.build_script_menu(mod_path)
        menu.append(('New Script ...', self.onAddNewScript, None))
        return menu

    #   save/load
#    def save_data(self, fid):
#        h2={}
#        if self.model is not None: h2["model"]=self.model.id
#        if self.sol is not None: h2["sol"]=self.sol.id
#        pickle.dump(h2, fid)

    def load_data(self, fid):
        h2 = pickle.load(fid)
        self.setvar("load_property", h2)

    def save_data2(self, data):
        def get_right_path(self, obj):
            if self.isdescendant(obj):
                return self.get_td_path(obj)
            else:
                return self.get_td_path(obj, True)
        check_ref(self)
        h2 = {}
#        if self.model is not None: h2["model"]=self.model.id
#        if self.sol is not None: h2["sol"]=self.sol.id
#        if self.param is not None: h2["param"]=self.param.id
        if self._model is not None:
            h2["model_path"] = get_right_path(self, self._model)
        if self._sol is not None:
            h2["sol_path"] = get_right_path(self, self._sol)
        if self._param is not None:
            h2["param_path"] = get_right_path(self, self._param)
        data['PySolver'] = (1, h2)

        data = super(PySolver, self).save_data2(data)
        return data

    def load_data2(self, data):
        h2 = data['PySolver'][1]
        self.setvar("load_property", h2)
        super(PySolver, self).load_data2(data)

    def init_after_load(self, olist, nlist):
        h2 = self.getvar("load_property")
        if "model" in h2:
            oid = h2["model"]
            self._model = self.resolve_olist_nlist_map(oid, olist, nlist)
        if "sol" in h2:
            oid = h2["sol"]
            self._sol = self.resolve_olist_nlist_map(oid, olist, nlist)
        if "param" in h2:
            oid = h2["param"]
            self._param = self.resolve_olist_nlist_map(oid, olist, nlist)

        if "model_path" in h2:
            self._model = self.resolve_td_path(h2["model_path"])
        if "sol_path" in h2:
            self._sol = self.resolve_td_path(h2["sol_path"])
        if "param_path" in h2:
            self._param = self.resolve_td_path(h2["param_path"])

        self.delvar("load_property")

    @property
    def model(self):
        return self._model

    @property
    def sol(self):
        return self._sol

    @property
    def param(self):
        return self._param


class BaseSolver(PySolver):
    '''
    this solver expose all menues
    so that user can setup whatever he/she wants
    '''

    def __init__(self, *args, **kargs):
        PySolver.__init__(self, *args, **kargs)
        self._queue = None    # queue used in solver script

    def tree_viewer_menu(self):
     # return MenuString, Handler, MenuImage
        return [('Run  (F9)', self.onRun, None),
                ('+Add Module',  None, None)] + \
            self.module_addon_list() + \
            [('!',  None, None),
                ('+Add Script',  None, None)] + \
            self.script_addon_list() + \
            [('!',  None, None),
                ('Setting...', self.onSetting, None),
                ('Initialize Solution', self.onInitSol, None),
                ('---', None, None)] + \
            super(PyCode, self).tree_viewer_menu()

    def load_default_script(self):
        pass

    def onRun(self, e=None):

        if not self.is_valid_setting():
            return
        val = self.init_sol()
        if not val:
            return
        self.solver_script.Run()
        if e is not None:
            e.Skip()
        return

    def onInitSol(self, e):
        if not cbook.ProxyAlive(self._model):
            ret = dialog.message(None,
                                 'Set working model in setting panel',
                                 'Model is not specified',
                                 )
            return

        val = self.init_sol()
#        if not cbook.ProxyAlive(self._param):
#            ret=dialog.message(None,
#                      'global parameter is not specified and was not copied',
#                      'Warning: No global parameter copied')
        ifigure.events.SendChangedEvent(self, e.GetEventObject())

    def is_valid_setting(self):
        return True

    def init_sol(self):
        return init_sol(self)

    def creat_sol(self):
        '''
        add a new sol folder
        '''
        root = self.get_root_parent()
        sol_exist = False
#        if not cbook.ProxyAlive(self._sol):
#           print "creating solution folder"
        root.set_psol(None)
        root = self.get_root_parent()
        root.onAddSol()
        psol = root.psol
        self._sol = psol
        psol.setvar("solver", self.get_full_path())
        return psol, sol_exist

    def get_target_model(self, model, solbase):
        modelbase = model.get_root_model()
        path = [model.name]
        while (model.get_full_path() !=
               modelbase.get_full_path()):
            model = model.get_parent()
            path = [model.name]+path

        sol = solbase
        for name in path:
            sol = sol.get_child(name=name)
        return sol

    def onSetting(self, e):
        list = onSettingList(self)
        app = wx.GetApp().TopWindow
        app.proj_tree_viewer.OpenPanel(list, self, 'setSetting')

    def setSetting(self, value):
        setSetting(self, value)

    def get_workers(self):
        return [w for name, w in self.get_children() if isinstance(w, PySol)]


class PyStdSolver(BaseSolver):
    ''' 
    standard solver is almost same as the base solver.
    it has an additional run scriopt already.
    '''

    def __init__(self, *args, **kargs):
        super(PyStdSolver, self).__init__(self, *args, **kargs)
        self._use_def_merger = True

    @classmethod
    def get_namebase(self):
        return 'stdsolver'

    def load_default_script(self):
        base_mod = 'ifigure.add_on.solver.script'
        mod_path = cbook.GetModuleDir(base_mod)
        file = os.path.join(mod_path, 'run_stdsolver.py')
        sc = PyScript()
        self.add_child('solver_script', sc)
        sc.import_script(file)
        self.add_default_mergesol_script()

    def add_default_mergesol_script(self):
        if self.has_child('merge_sol'):
            return
        base_mod = 'ifigure.add_on.solver.script'
        mod_path = cbook.GetModuleDir(base_mod)
        file = os.path.join(mod_path, 'merge_sol_std.py')
        sc = PyScript()
        self.add_child('merge_sol', sc)
        sc._name_readonly = True
        sc.import_script(file)
#        sc.setvar("task", 2)

    def tree_viewer_menu(self):
        # return MenuString, Handler, MenuImage
        l = [('Run  (F9)', self.onRun, None),
             ('Setting...', self.onSetting, None),
             ('Initialize Solution', self.onInitSol, None), ]
        return l + [('---', None, None), ] + \
            super(PyCode, self).tree_viewer_menu()

    def onAbort(self, evt):
        self._queue.put('abort job')

    def onRun(self, e=None):
        if not self.is_valid_setting():
            return

        if self._sol:
            ret = dialog.message(
                message="Solution folder exists. Do you want to erase it?\n (No: New Sol, Yes:Overwrite, Cancel:Stop)", style=5)
            if ret == 'cancel':
                return
            if ret == 'yes':
                self._sol.destroy()
            ifigure.events.SendChangedEvent(self, wx.GetApp().TopWindow)

        self.init_sol()
        self.solver_script.Run()
        if e is not None:
            e.Skip()
        return

    def init_sol(self, sol=None):
        if sol is None:
            psol, sol_exist = self.creat_sol()
        else:
            print("error: sol is supposed to be None here")
            return

        for x in [x for x in self.walk_tree() if isinstance(x, PySol)]:
            x.destroy()

        ps = PySol()
        self.add_child('worker', ps)
        ps.add_child('parameters', PyParam())

        ifigure.events.SendChangedEvent(self)
        return True

    def onSetting(self, e):
        list = onSettingList(self)
        list = list + [
            [None,  self._use_def_merger, 3,
                {"text": "Use standard merger", "expand": True}], ]
        app = wx.GetApp().TopWindow
        app.proj_tree_viewer.OpenPanel(list, self, 'setSetting')

    def setSetting(self, value):
        setSetting(self, value)
        self._use_def_merger = value[4]

    def save_data2(self, data):
        vars = ('_use_def_merger', )
        h2 = {key: getattr(self, key) for key in vars}
        data['PyStdSolover'] = (1, h2)
        data = super(PyStdSolver, self).save_data2(data)
        return data

    def load_data2(self, data):
        super(PyStdSolver, self).load_data2(data)
        if 'PyStdSolver' in data:
            h2 = data['PyStdSolver'][1]
            for key in h2:
                setattr(self, key, h2[key])


class PyParametric(BaseSolver):
    ''' 
    parametric solver is for parametric scan.
    it has an addional setting field _param_names
    and modifier module, which modify the model
    based on parameter
    '''

    def __init__(self, *args, **kargs):
        super(PyParametric, self).__init__(self, *args, **kargs)
        self._pname = None
        self._pvalue = None
        self._num_worker = 2
        self._expand_itertools_product = False
        self._expand_lhs = False
        self._expand_lhs_parmas = [10, "center"]
        self._use_def_merger = False
        self.setvar("param name", ("name1", "name2"))
        self.setvar("param value", ((1, 1), (1, 2), (2, 1), (2, 2)))
#        self.setvar("use thread", False)
    @classmethod
    def get_namebase(self):
        return 'parametric'

    def tree_viewer_menu(self):
        # return MenuString, Handler, MenuImage
        l = [('Run  (F9)', self.onRun, None),
             ('Setting...', self.onSetting, None),
             ('Initialize Solution', self.onInitSol, None),
             ('Solve All Cases', self.onSolveAllCases, None),
             ('Solve Cases...', self.onSolveCases, None), ]
        if self._queue is not None:
            l.append(('Abort', self.onAbort, None))
        return l + [('---', None, None), ] + \
            super(PyCode, self).tree_viewer_menu()

    def onAbort(self, evt):
        self._queue.put('abort job')

    def load_default_script(self):
        base_mod = 'ifigure.add_on.solver.script'
        mod_path = cbook.GetModuleDir(base_mod)
        file = os.path.join(mod_path, 'run_parametric.py')
        sc = PyScript()
        self.add_child('solver_script', sc)
        sc._name_readonly = True
        sc.import_script(file)
        sc.setvar("task", 2)
        self.add_default_mergesol_script()

    def add_default_mergesol_script(self):
        if self.has_child('merge_sol'):
            return
        base_mod = 'ifigure.add_on.solver.script'
        mod_path = cbook.GetModuleDir(base_mod)
        file = os.path.join(mod_path, 'merge_sol_parametric.py')
        sc = PyScript()
        self.add_child('merge_sol', sc)
        sc._name_readonly = True
        sc.import_script(file)
#        sc.setvar("task", 2)

    def is_valid_setting(self):
        if (self._pname is None or
                self._pvalue is None):
            return False
        return True

    def init_sol(self, sol=None):
        if not self.is_valid_setting():
            return
        if sol is None:
            psol, sol_exist = self.creat_sol()
        else:
            print("error: sol is supposed to be None here")
            return

        vname = self._pname
        value = self._pvalue
        use_worker = self._num_worker > 0
        name_base = 'case'

        for name, child in psol.get_children():
            child.destroy()

        parent = psol
#        print [x for x in self.walk_tree() if isinstance(x, PySol)]
        for x in [x for x in self.walk_tree() if isinstance(x, PySol)]:
            x.destroy()
        if use_worker:
            #            for x in range(self._num_worker):
            #                ps = PySol()
            #                self.add_child('worker'+str(x), ps)
            value = self._pvalue[:self._num_worker]
            name_base = 'worker'
            parent = self

        i = 0

        app = wx.GetApp().TopWindow
        pgb = dialog.progressbar(app,
                                 'prepareing '+str(len(value)) +
                                 ' sol folders',
                                 'initialize sol', len(value))

        check = True
        for i, v in enumerate(value):
            ps = PySol()
            parent.add_child(name_base + str(i), ps)
            ps.setvar("name", vname)
            ps.setvar("value", v)
            if i == 0:
                flag, tmpdir = init_sol(self, sol=ps, mode=1)
                check = check and flag
            else:
                check = check and init_sol(self, sol=ps, mode=2,
                                           tmpdir=tmpdir)
            pgb.Update(i)
        shutil.rmtree(tmpdir+'_bk')
        pgb.Destroy()
        ifigure.events.SendChangedEvent(self)
        return check

    def print_cases(self):
        print(self._pvalue)

    def onRun(self, evt):
        if not self.is_valid_setting():
            return
        if self._sol:
            ret = dialog.message(
                message="Solution folder exists. Do you want to erase it?\n(No: New Sol, Yes:Overwrite, Cancel:Stop)", style=5)
            if ret == 'cancel':
                return
            if ret == 'yes':
                self._sol.destroy()
            ifigure.events.SendChangedEvent(self, wx.GetApp().TopWindow)

        self.init_sol()
        sc = self.solver_script
        self._run_cases = [True]*len(self._pvalue)
        sc.setvar("task", 2)
        sc.do_run()
        sc.setvar("task", 2)
        if evt is not None:
            evt.Skip()

    def onApplyModifier(self, evt):
        sc = self.solver_script
        sc.setvar("task", 1)
        sc.do_run()
        sc.setvar("task", 2)

    def onSolveAllCases(self, evt):
        sc = self.solver_script
        self._run_cases = [True]*len(self._pvalue)
        sc.setvar("task", 2)
        sc.do_run()
        sc.setvar("task", 2)

    def onSolveCases(self, evt):
        shellvar = self.get_root_parent().app.shell.lvar
        l = [["", "Enter Case Index", 2],
             ["Index", '[1]', 100, {'ns': shellvar}], ]
        app = wx.GetApp().TopWindow
        app.proj_tree_viewer.OpenPanel(l, self, 'run_cases')

    def run_cases(self, value=None, index=None):
        self._run_cases = [False]*len(self._pvalue)
        if value is not None:
            for x in value[1][1]:
                self._run_cases[x] = True
        elif index is not None:
            for x in index:
                self._run_cases[x] = True

        sc = self.solver_script
        sc.setvar("task", 2)
        sc.do_run()
        sc.setvar("task", 2)

    def apply_modifier(self, modifier, case_xx, i):
        sols = self.get_pysol(self._sol)
        num = len(sols)
#        pgb = dialog.progressbar(self.get_root_parent().app,
#                    'prepareing '+str(num)+ ' sol folders',
#                    'initialize sol', num)

        vname = self._pname
        value = self._pvalue

#        for i, case_xx in enumerate(sols):
        case_xx.setvar("name", vname)
        case_xx.setvar("value", value[i])
        n = case_xx.getvar("name")
        if (not isinstance(n, list) and
                not isinstance(n, tuple)):
            n = (n,)
        v = case_xx.getvar("value")
        if (not isinstance(v, list) and
                not isinstance(v, tuple)):
            v = (v,)
        modifier(case_xx, n, v)
#        pgb.Update(i)
#        pgb.Destroy()

    def get_expand_setting_value(self):
        expand = 'Expand...'
        if self._expand_itertools_product:
            expand = 'Product'
        if self._expand_lhs:
            expand = 'LatinHyperCube'
        data = [expand,
                [None,],
                [None,],
                [None, str(self._expand_lhs_parmas[0]), self._expand_lhs_parmas[1]],]
        return data
    
    def call_lhs(self, params, xlimits):
        num = int(params[0])
        method = params[1]
        
        from smt.sampling_methods import LHS

        try:
            xlimits = np.array(list(xlimits))            
            sampling = LHS(xlimits=xlimits, criterion=method)            
            x = sampling(num)
            
        except BaseException:
            import traceback
            traceback.print_exc()

            print("Error!")
            print("LHS was called with following parameters...")
            print(num, method, xlimits)
            return []
        ## I need to transform it to tuple...;D
        return tuple([tuple(xx) for xx in x])

    def onSetting(self, e):
        list = onSettingList(self)
        txt1 = str(self.getvar("param name"))
        txt2 = str(self.getvar("param value"))
        shellvar = self.get_root_parent().app.shell.lvar

        expand_values = self.get_expand_setting_value()
        print("value here", expand_values)
        ss1 = {"style": wx.CB_READONLY,
               "choices":["center", "maximin", "centermaximin", "correlation",
                          "c", "m", "cm", "corr", "ese"]}
        list = list + [
            ["Parameter Names", txt1, 100, {'ns': shellvar}],
            ["Parameter Values", txt2, 100, {'ns': shellvar}],
            [None,  expand_values, 34, [{'text': "",
                                'call_fit': False, 
                                'choices': ['Expand...',
                                            'Product',
                                            'LatinHyperCube',],},
                     {'elp':[(None, "Parameters are used As-Is", 2, None)]},
                     {'elp':[(None, "Parameters are expanded using Itertools.product", 2, None)]},
                     {'elp':[(None, "Parameters are expanded using smt.sampling_methods.LHS", 2, None),
                             ("#points", "10", 0, None),
                             ("method", "center", 4, ss1)
                      ]},]],
            #[None,  self._expand_itertools_product, 3,
            # {"text": "Expand parameters by itertool.product", "expand": True}],
            [None,  self._use_def_merger, 3,
             {"text": "Use standard merger", "expand": True}],
            [None,  self._num_worker > 0, 3,
             {"text": "Use workers", "expand": True}],
            ["Number of workers", str(abs(self._num_worker)), 0], ]
#                ["Use multiple thread", txt3, 1 ]]
        app = wx.GetApp().TopWindow

        style = app.proj_tree_viewer.get_defaultpanelstyle()
        style = wx.RESIZE_BORDER|style
        app.proj_tree_viewer.OpenPanel(list, self, 'setSetting',
                                       title='Parametric sweep',
                                       style=style)

    def setSetting(self, value):
        setSetting(self, value)
        self.setvar("param name", str(value[4][0]))
        self.setvar("param value", str(value[5][0]))

        self._pname = value[4][1]

        self._expand_itertools_product = False
        self._expand_lhs = False
        
        if value[6][0] == 'Product':
            import itertools
            self._expand_itertools_product = True
            self._pvalue = tuple(itertools.product(*(value[5][1])))
        elif value[6][0] == 'LatinHyperCube':
            import itertools
            self._expand_lhs = True
            self._expand_lhs_parmas = value[6][3][1:]
            self._pvalue = self.call_lhs(value[6][3][1:], value[5][1])
        else:
            self._pvalue = value[5][1]
        self._num_worker = abs(
            int(value[9])) if value[8] else -abs(int(value[8]))
        self._use_def_merger = value[7]
#        if value[6] == 'on':
#            use_thread = self.setvar("use thread", True)
#        else:
#            use_thread = self.setvar("use thread", False)

    def save_data2(self, data):
        vars = ('_expand_itertools_product', '_pname',
                '_pvalue', '_num_worker', '_use_def_merger')
        h2 = {key: getattr(self, key) for key in vars}
        data['PyParametric'] = (1, h2)
        data = super(PyParametric, self).save_data2(data)
        return data

    def load_data2(self, data):
        super(PyParametric, self).load_data2(data)
        if 'PyParametric' in data:
            h2 = data['PyParametric'][1]
            for key in h2:
                setattr(self, key, h2[key])


class PyOptimizer(PyStdSolver):
    def __init__(self, *args, **kargs):
        super(PyOptimizer, self).__init__(*args, **kargs)
        self.setvar("param name", ("name1", "name2"))
        self.setvar("init_value", (1, 1))
        self.setvar("opt_kwrds", {})
        self._pname = ('name1', 'name2')
        self._init_value = (1, 1)
        self._cost = '.cost'

    @classmethod
    def get_namebase(self):
        return 'optimizer'

    def load_default_script(self):
        base_mod = 'ifigure.add_on.solver.script'
        mod_path = cbook.GetModuleDir(base_mod)
        file = os.path.join(mod_path, 'run_optimizer.py')
        sc = PyScript()
        self.add_child('solver_script', sc)
        sc.import_script(file)
        self.add_default_mergesol_script()
        self.add_default_cost_script()

    def add_default_cost_script(self):
        if self.has_child('cost'):
            return
        base_mod = 'ifigure.add_on.solver.script'
        mod_path = cbook.GetModuleDir(base_mod)
        file = os.path.join(mod_path, 'cost_optimizer.py')
        sc = PyScript()
        self.add_child('cost', sc)
        sc._name_readonly = True
        sc.import_script(file)
#        sc.setvar("task", 2)

    def tree_viewer_menu(self):
        # return MenuString, Handler, MenuImage
        l = super(PyOptimizer, self).tree_viewer_menu()
        if self._queue is not None:
            l.insert(3, ('Abort', self.onAbort, None))
        return l

    def onSetting(self, e):
        list = onSettingList(self)
        txt1 = str(self.getvar("param name"))
        txt2 = str(self.getvar("init_value"))
        txt3 = str(self.getvar("opt_kwrds"))
        shellvar = self.get_root_parent().app.shell.lvar

        list = list + [
            ["Parameter Names", txt1, 100, {'ns': shellvar}],
            ["Initial Values", txt2, 100, {'ns': shellvar}],
            ["Optimize kwrds", txt3, 100, {'ns': shellvar}],
            ["Cost", '.cost', 0],
            [None,  self._use_def_merger, 3,
                {"text": "Use standard merger", "expand": True}], ]

        app = wx.GetApp().TopWindow
        app.proj_tree_viewer.OpenPanel(list, self, 'setSetting')

    def setSetting(self, value):
        setSetting(self, value)
        self.setvar("param name", str(value[4][0]))
        self.setvar("init_value", str(value[5][0]))
        self.setvar("opt_kwrds", str(value[6][0]))

        self._pname = value[4][1]
        self._init_value = value[5][1]
        self._opt_kargs = value[6][1]
        self._cost = str(value[7])
        self._use_def_merger = value[8]

    def onAbort(self, evt):
        self._queue.put('abort job')

    def save_data2(self, data):
        vars = ('_pname', '_cost',
                '_opt_kargs', '_init_value', '_use_def_merger')
        h2 = {key: getattr(self, key) for key in vars}
        data['PyOptimizer'] = (1, h2)
        data = super(PyOptimizer, self).save_data2(data)
        return data

    def load_data2(self, data):
        super(PyOptimizer, self).load_data2(data)
        if 'PyOptimizer' in data:
            h2 = data['PyOptimizer'][1]
            for key in h2:
                setattr(self, key, h2[key])



from __future__ import print_function
import ifigure.utils.cbook as cbook
import ifigure
import os
import sys
import weakref
import logging
import threading
import ifigure.utils.pickle_wrapper as pickle
from ifigure.mto.treedict import TopTreeDict, TreeDict
from ifigure.mto.py_code import PyParam, AbsModuleContainer, AbsScriptContainer, AbsFileContainer
from ifigure.mto.py_code import PyModel
from ifigure.mto.py_solver import PySolver
from ifigure.mto.py_code import PySol
from ifigure.mto.fig_book import FigBook

import ifigure.widgets.dialog as dialog

import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('ProjectTop')

lock = threading.Lock()


class PySetting(TreeDict, AbsModuleContainer, AbsScriptContainer):
    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx = cbook.LoadImageFile(path, 'setting.png')
        return [idx]

    @classmethod
    def get_namebase(self):
        return 'setting'

    @classmethod
    def isPySetting(self):
        return True

    def tree_viewer_menu(self):
        from ifigure.mto.py_code import PyParam
        has_pyparam = False
        for name, child in self.get_children():
            if isinstance(child, PyParam):
                has_pyparam = True
                break
        if has_pyparam:
            v = [('-Add Parameters', self.onAddParam, None), ]
        else:
            v = [('Add Parameters', self.onAddParam, None), ]
     # return MenuString, Handler, MenuImage
        return v + [('+Add Module',  None, None)] + \
            self.module_addon_list() + \
            [('!',  None, None),
             ('+Add Script',  None, None)] + \
            self.script_addon_list() + \
            [('!',  None, None),
             ('---',  None, None)] +     \
            super(PySetting, self).tree_viewer_menu()

    def module_addon_list(self):
        base_mod = 'ifigure.add_on.setting.module'
        mod_path = cbook.GetModuleDir(base_mod)
        menu = self.build_module_menu(mod_path)
        menu.append(('New Module...', self.onAddNewModule, None))
        return menu

    def script_addon_list(self):
        base_mod = 'ifigure.add_on.setting.script'
        mod_path = cbook.GetModuleDir(base_mod)
        menu = self.build_module_menu(mod_path)
        menu.append(('New Script...', self.onAddNewScript, None))
        return menu

    def onAddParam(self, evt):
        obj = PyParam()
        name = 'parameters'
        self.add_child(name, obj)
        evt.Skip()


class ProjectTop(TopTreeDict, AbsScriptContainer, AbsFileContainer):
    '''
    root object of treedict (data tree object)
    '''

    def __init__(self, src=None):
        #
        #  variable to specify "present" subtrees.
        #
        self._pbook = None
        self._pmodel = None
        self._psetting = None
        self._psolver = None
        self._psol = None
        self._saved = False

        super(ProjectTop, self).__init__(src=src)
        self._name = 'proj'

        if src is None:
            setting = PySetting()
            self.add_child('setting', setting)
            param = PyParam()
            setting.add_child('parameters', param)

    def set_saved(self, value):
        self._saved = value
#         import traceback
#         traceback.print_stack()

    def get_saved(self):
        return self._saved

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx = cbook.LoadImageFile(path, 'sun.png')
        return [idx]

    def write_log(self, *args):
        lock.acquire()
        name = os.path.join(self.eval("wdir"), 'log')
        fid = open(name, 'a+')
        import datetime
        import threading
        txt1 = '::'.join([datetime.datetime.now().strftime(
            "%Y_%m_%d_%H_%M"), threading.current_thread().name]) + '\n'
        fid.write(txt1)
        fid.write('\n'.join(['   '+str(x) for x in args]))
        fid.write('\n')
        fid.close()
        lock.release()

    def clear_log(self):
        name = os.path.join(self.eval("wdir"), 'log')
        fid = open(name, 'w')
        fid.write('')
        fid.close()

    def show_log(self):
        name = os.path.join(self.eval("wdir"), 'log')
        fid = open(name, 'r')
        print(''.join(fid.readlines()))
        fid.close()

    def tree_viewer_menu(self):
        # return MenuString, Handler, MenuImage
        from ifigure.mto.hg_support import has_hg

        menu = [
            ('+Add...', None, None),
            ('Book',   self.onAddBook, None),
            ('Text',   self.onAddNewText, None),
            ('+Script...',  None, None), ] + \
            self.script_template_list() + \
            [('!',  None, None),
             ('+Modeling...', None, None),
             ('Setting', self.onAddSetting, None),
             ('Model', self.onAddModel, None),
             ('Solver',   self.onAddSolver, None),
             ('Solution',   self.onAddSol, None),
             ('!', None, None),
             ('Other File...',   self.onAddOther, None),
             ('!', None, None),
             ('Setting...',   self.onSetting, None),
             ('---',   None, None), ]

        if has_hg:
            menu.append(('Check Package Updates',
                         self.onCheckPackageUpdate, None))
        menu.append(('+Import Subtree', None, None))
        mm = [('From File...', self.onLoadSubTree, None), ]
        if has_hg:
            mm.append(('From Repository...', self.onLoadSubTreeHG,
                       None))
        if hasattr(self, 'onAddExtFolder'):
            mm.append(('From Folder...', self.onAddExtFolder, None))
        menu.extend(mm)
        menu.append(('!', None, None))
        menu.extend([('Export All Subtree', self.onExportAllSubTree, None),
                     ('New Folder...', self.onAddFolder, None)])

        return menu

    @classmethod
    def isProjectTop(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'proj'

    def destroy(self, clean_owndir=True):
        self._pbook = None
        self._pmodel = None
        self._psetting = None
        self._psolver = None
        self._psol = None
        super(ProjectTop, self).destroy(clean_owndir=clean_owndir)

    def onCheckPackageUpdate(self, e):
        from ifigure.mto.hg_support import hg_check_all_incoming_outgoing, hg_verify_all_repo
        broken_repo = hg_verify_all_repo(self)
        if len(broken_repo) != 0:
            txt = ['Some repo are broken\n'] + \
                '\n'.join(['  ' + str(o) for o in broken_repo])
            ret = dialog.message(e.GetEventObject(), txt, 'Broken Repo', 0)
            return
        updated_obj, newer_obj, both_obj = hg_check_all_incoming_outgoing(self)
        #app = wx.GetApp().TopWindow
        # app.proj_tree_viewer.update_widget()#_request2()

        labels = ['pull  ' + obj.get_full_path() + '(local:'+str(rev) + ' <<-- repo' + str(rev2) + ')' for obj, l, rev, rev2 in updated_obj] + \
            ['push ' + obj.get_full_path() + '(local:'+str(rev) + ' -->> repo: ' +
             str(rev2) + ')' for obj, l, rev, rev2 in newer_obj]
        if len(labels) == 0:
            ret = dialog.message(e.GetEventObject(),
                                 'No update is found', 'No Update', 0)
            return
        # print labels, e.GetEventObject()
        from ifigure.utils.checkbox_panel import checkbox_panel
        value = checkbox_panel(e.GetEventObject(), col=1, labels=labels,
                               values=[True]*len(updated_obj) +
                               [False]*len(newer_obj),
                               message='Select packages to update',
                               title='Package updage', blabel='Update')

        if value is None:
            return
        from ifigure.mto.hg_support import update_to_latest
        for i, v in enumerate(value):
            label, flag = v
            if flag:
                if i < len(updated_obj):
                    new_rev, new_obj = update_to_latest(updated_obj[i][0])
                    dprint1(str(new_obj)+'was updated to ' + str(new_rev))
                else:
                    new_rev, url = update_to_latest(
                        newer_obj[i-len(updated_obj)][0], rev=True)
                    dprint1('repo('+url+') was updated to ' + str(new_rev))
        # dprint1('skipping')
        e.Skip()  # this will update widget

    def onSetting(self, e):
        id = e.GetEventObject()
        app = id.GetTopLevelParent()

        setting = ''
        model = ''
        solver = ''
        sol = ''
        book = ''

        if cbook.ProxyAlive(self._psetting):
            setting = self._psetting.get_full_path()
        if cbook.ProxyAlive(self._pmodel):
            model = self._pmodel.get_full_path()
        if cbook.ProxyAlive(self._psolver):
            solver = self._psolver.get_full_path()
        if cbook.ProxyAlive(self._psol):
            sol = self._psol.get_full_path()
        if cbook.ProxyAlive(self._pbook):
            book = self._pbook.get_full_path()

        list = [["", "Project Setting", 2],
                ["Setting", setting, 0],
                ["Model", model, 0],
                ["Solver", solver, 0],
                ["Solution", sol, 0],
                ["Book",  book, 0]]

        app.proj_tree_viewer.OpenPanel(list, self, 'setSetting')

    def setSetting(self, value):
        logging.basicConfig(level=logging.DEBUG)

        root = self.get_root_parent()
        exec(root._name + ' = root')
        if len(value[1]) != 0:
            try:
                b = eval(value[1])
                b.isPySetting()
                self._psetting = weakref.proxy(b)
            except Exception:
                self._psetting = None
                logging.exception("")
        if len(value[2]) != 0:
            try:
                b = eval(value[2])
                b.isPyModel()
                self._pmodel = weakref.proxy(b)
            except Exception:
                self._pmodel = None
                logging.exception("")
        if len(value[3]) != 0:
            try:
                b = eval(value[3])
                b.isPySolver()
                self._psolver = weakref.proxy(b)
            except Exception:
                self._psolver = None
                logging.exception("")
        if len(value[4]) != 0:
            try:
                b = eval(value[4])
                b.isPySol()
                self._psol = weakref.proxy(b)
            except Exception:
                self._psol = None
                logging.exception("")
        if len(value[5]) != 0:
            try:
                b = eval(value[5])
                b.isFigBook()
                self._pbook = weakref.proxy(b)
            except Exception:
                self._pbook = None
                logging.exception("")

    def set_psetting(self, a):
        self._psetting = None
        if a is not None:
            self._psetting = weakref.proxy(a)

    def set_pmodel(self, a):
        self._pmodel = None
        if a is not None:
            self._pmodel = weakref.proxy(a)

    def set_pbook(self, a):
        self._pbook = None
        if a is not None:
            self._pbook = weakref.proxy(a)

    def set_psol(self, a):
        self._psol = None
        if a is not None:
            self._psol = weakref.proxy(a)

    def set_psolver(self, a):
        self._psolver = None
        if a is not None:
            self._psolver = weakref.proxy(a)

    def num_book(self):
        return self._countSomething('isFigBook')

    def num_model(self):
        return self._countSomething('isPyModel')

    def num_solver(self):
        return self._countSomething('isPySolver')

    def num_sol(self):
        return self._countSomething('isPySol')

    def num_setting(self):
        return self._countSomething('isPySetting')

    @property
    def psetting(self):
        return self._psetting

    @property
    def pmodel(self):
        return self._pmodel

    @property
    def pbook(self):
        return self._pbook

    @property
    def psolver(self):
        return self._psolver

    @property
    def psol(self):
        return self._psol

    #
    #  Pulldown Menu
    #
    def onAddNewScript(self, e=None, **kargs):
        child = AbsScriptContainer.onAddNewScript(self, e, **kargs)
        if child is not None:
            child._simplemenu = True
        return child

    def onAddSetting(self, e=None):
        setting = PySetting()
        self._onAddSomething(setting,
                             ['isPySolver', 'isFigBook', 'isPyModel', 'isPySol'])
        if not cbook.ProxyAlive(self._psetting):
            self._psetting = weakref.proxy(setting)
        if e is not None:
            e.Skip()
        return setting

    def onAddModel(self, e=None):
        model = PyModel()
        basename = model.get_namebase()
        name = self.get_next_name(basename)
        if e is not None:
            id = e.GetEventObject()
            app = id.GetTopLevelParent()
            ret, new_name = dialog.textentry(app,
                                             "Enter the name of new model",
                                             "Add Model", name)
            if not ret:
                return
        else:
            new_name = name
        self._onAddSomething(model, ['isPySolver', 'isFigBook', 'isPySol'],
                             name=new_name)
        if not cbook.ProxyAlive(self._pmodel):
            self._pmodel = weakref.proxy(model)
        if e is not None:
            e.Skip()
        return model

    def onAddSolver(self, e=None):
        solver = PySolver()
        self._onAddSomething(solver, ['isFigBook', 'isPySol'])
        if not cbook.ProxyAlive(self._psolver):
            self._psolver = weakref.proxy(solver)
        if e is not None:
            e.Skip()
        return solver

    def onAddSol(self, e=None):
        sol = PySol()
        self._onAddSomething(sol, ['isFigBook'])
        if not cbook.ProxyAlive(self._psol):
            self._psol = weakref.proxy(sol)
        if e is not None:
            e.Skip()
        return sol

    def onAddBook(self, e=None, basename=None):
        book = FigBook()
        self._onAddSomething(book, basename=basename)
        if not cbook.ProxyAlive(self._pbook):
            self._pbook = weakref.proxy(book)
        if e is not None:
            e.Skip()
        return book

    def onAddExtFolder(self, evt):
        from ifigure.widgets.dialog import readdir
        dir = readdir(parent=evt.GetEventObject(),
                      message='Select external folder to import')
        if dir == '':
            return
        self.add_extfolder(dir)
        if evt is not None:
            evt.Skip()

    def _onAddSomething(self, item, methods=None, basename=None, name=None):
        if name is None:
            if basename is None:
                basename = item.get_namebase()
            name = self.get_next_name(basename)
        idx = self.add_child(name, item)
        if methods is None:
            return

        idx2 = 0
        for name, child in self.get_children():
            for m in methods:
                if getattr(child, m, None) is not None:
                    self.move_child(idx, idx2)
                    return
            idx2 = idx2+1

    def _countSomething(self, method=None):
        c = 0
        for name, child in self.get_children():
            if getattr(child, method, None) is not None:
                c = c+1
        return c

    def add_folder(self, name):
        from ifigure.mto.py_code import PyFolder
        return self.add_childobject(PyFolder, name)

    def add_book(self, name):
        from ifigure.mto.fig_book import FigBook
        return self.add_childobject(FigBook, name)

    def add_script(self, name):
        from ifigure.mto.py_script import PyScript
        return self.add_childobject(PyScript, name)

    def add_data(self, name):
        from ifigure.mto.py_code import PyData
        return self.add_childobject(PyData, name)

    def add_model(self, name):
        '''
        add a new model
        '''
        from ifigure.mto.py_code import PyModel
        return self.add_childobject(PyModel, name)

    def add_file(self, name):
        from ifigure.mto.py_file import PyFile
        return self.add_childobject(PyFile, name)

    def add_text(self, name):
        from ifigure.mto.py_file import PyText
        return self.add_childobject(PyText, name)

    def add_solver(self, name):
        return self.add_childobject(PySolver, name)

    def reorder_top_level(self):
        d = {'book': [],
             'model': [],
             'solver': [],
             'sol': [],
             'setting': [],
             'others': []}
        for name, child in self.get_children():
            if isinstance(child, PySetting):
                d['setting'].append(child)
            elif isinstance(child, FigBook):
                d['book'].append(child)
            elif isinstance(child, PyModel):
                d['model'].append(child)
            elif isinstance(child, PySol):
                d['sol'].append(child)
            elif isinstance(child, PySolver):
                d['solver'].append(child)
            else:
                d['others'].append(child)
        self._d = d['setting']+d['model'] + \
            d['solver']+d['sol']+d['book']+d['others']

    def load_subtree(self, filename, keep_zorder=True, message=None,
                     compress=False, usetar=True):

        ret = super(ProjectTop, self).load_subtree(filename,
                                                   keep_zorder=keep_zorder,
                                                   message=message,
                                                   compress=compress,
                                                   usetar=usetar)
        self.reorder_top_level()
        return ret

    def save_data(self, fid):
        h2 = {}
        if cbook.ProxyAlive(self._psetting):
            h2["isetting"] = self._psetting.id
        if cbook.ProxyAlive(self._pmodel):
            h2["imodel"] = self._pmodel.id
        if cbook.ProxyAlive(self._psolver):
            h2["isolver"] = self._psolver.id
        if cbook.ProxyAlive(self._pbook):
            h2["ibook"] = self._pbook.id
        if cbook.ProxyAlive(self._psol):
            h2["isol"] = self._psol.id
        pickle.dump(h2, fid)

    def load_data(self, fid):
        h2 = pickle.load(fid)
        self.setvar("load_property", h2)

    def save_data2(self, data):
        def get_right_path(self, obj):
            if self.isdescendant(obj):
                return self.get_td_path(obj)
            else:
                return self.get_td_path(obj, True)

        h2 = {}
#         if cbook.ProxyAlive(self._psetting): h2["isetting"]=self._psetting.id
#         if cbook.ProxyAlive(self._pmodel): h2["imodel"]=self._pmodel.id
#         if cbook.ProxyAlive(self._psolver): h2["isolver"]=self._psolver.id
#         if cbook.ProxyAlive(self._pbook): h2["ibook"]=self._pbook.id
#         if cbook.ProxyAlive(self._psol): h2["isol"]=self._psol.id
        if cbook.ProxyAlive(self._psetting):
            h2["isetting_path"] = get_right_path(self, self._psetting)
        if cbook.ProxyAlive(self._pmodel):
            h2["imodel_path"] = get_right_path(self, self._pmodel)
        if cbook.ProxyAlive(self._psolver):
            h2["isolver_path"] = get_right_path(self, self._psolver)
        if cbook.ProxyAlive(self._pbook):
            h2["ibook_path"] = get_right_path(self, self._pbook)
        if cbook.ProxyAlive(self._psol):
            h2["isol_path"] = get_right_path(self, self._psol)
        data['ProjectTop'] = (1, h2)
        return data

    def load_data2(self, data):
        self.setvar("load_property", data['ProjectTop'][1])

    def onExportAllSubTree(self, e):
        from ifigure.mto.py_code import PyModel

        for child in self.walk_tree():
            #             if isinstance(child, PyModel):
            if (child.hasvar('subtree_path') and
                    os.path.exists(child.getvar('subtree_path'))):
                path = child.getvar('subtree_path')
                print('exporting subtree to '+path)
                child.save_subtree(path)

    def init_after_load(self, olist, nlist):
        #
        #  init_after_load is initialization
        #  after loading entire tree
        #
        #  note: self.get_child would fail
        #        if it is issued in load_data,
        #        since children are not yet
        #        loaded.
        #
        h2 = self.getvar("load_property")

        if "isetting" in h2:
            oid = h2["isetting"]
            self._psetting = self.resolve_olist_nlist_map(oid, olist, nlist)
        if "imodel" in h2:
            oid = h2["imodel"]
            self._pmodel = self.resolve_olist_nlist_map(oid, olist, nlist)
        if "isolver" in h2:
            oid = h2["isolver"]
            self._psolver = self.resolve_olist_nlist_map(oid, olist, nlist)
        if "ibook" in h2:
            oid = h2["ibook"]
            self._pbook = self.resolve_olist_nlist_map(oid, olist, nlist)
        if "isol" in h2:
            oid = h2["isol"]
            self._psol = self.resolve_olist_nlist_map(oid, olist, nlist)

        if "isetting_path" in h2:
            self._psetting = self.resolve_td_path(h2["isetting_path"])
        if "imodel_path" in h2:
            self._pmodel = self.resolve_td_path(h2["imodel_path"])
        if "isolver_path" in h2:
            self._psolver = self.resolve_td_path(h2["isolver_path"])
        if "ibook_path" in h2:
            self._pbook = self.resolve_td_path(h2["ibook_path"])
        if "isol_path" in h2:
            self._psol = self.resolve_td_path(h2["isol_path"])

        self.delvar("load_property")

        if self._pbook is None:
            print(("??? self._pbook is None", "adjusting ..."))
            if self.num_book() == 0:
                self.onAddBook()
            else:
                for name, child in self.get_children():
                    if isinstance(child, FigBook):
                        self.set_pbook(child)
            print(self._pbook)



from __future__ import print_function
import Pyro4
import subprocess
import shlex
import weakref
import socket
import logging
import time


def PickUnusedPort():
    '''
    a routeint to le a system to pick an unused port.
    this recipe was found in active state web site
    '''

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(('localhost', 0))
    addr, port = s.getsockname()
    s.close()
    return port


def confirm_connection(host, port):
    retry = 0
    max_retry = 15
    retry = 0
    wait = 2

    if host == 'localhost':
        host1 = ''
    else:
        host1 = host
    arg = (host1, port)

    while retry < max_retry:
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect(arg)
        except Exception:
            #          logging.exception("connection failed")
            retry = retry+1
            time.sleep(wait)
            continue
        break
    s.close()
    if retry == max_retry:
        print("connection failed")
        return False
    print("connection established")
    return True


class PyroSSH(object):
    def __init__(self, *args, **kargs):
        self.process = []
        self.proxy = []
        return super(PyroSSH, self).__init__()

    def nslookup(self, host, name, port=9090):
        lport = PickUnusedPort()  # unused local port
        command = 'ssh -N '+str(host) + ' -L ' +  \
            str(lport)+':' + str(host)+':' + \
            str(port)  # + ' & echo $!'
#        p=subprocess.call(shlex.split(command), shell=True)
        print(command)
        p = subprocess.Popen(shlex.split(command),
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT)

        confirm_connection('localhost', lport)
        try:
            nameserver = Pyro4.locateNS(port=lport)
            uri = nameserver.lookup(name)
        except Exception:
            return None
#           logging.exception("nslookup failed")
#           p.terminate()
#           return None
        print(('killing process', p.pid))
        p.terminate()
        p.kill()
        return uri

    def open_proxy(self, uri):
        lport = PickUnusedPort()  # unused local port
        command = 'ssh -N  '+str(uri.host) + ' -L ' +  \
            str(lport)+':' + str(uri.host)+':' + \
            str(uri.port)
        print(command)
        p = subprocess.Popen(shlex.split(command))
        confirm_connection('localhost', lport)

        uri0 = Pyro4.URI("PYRO:"+uri.object+"@localhost:"+str(lport))
        proxy = Pyro4.Proxy(uri0)
        self.process.append(p)
        self.proxy.append(proxy)
        return proxy

    def close_proxy(self, proxy):
        if self.proxy.count(proxy) != 0:
            idx = self.proxy.index(proxy)
            p = self.process[idx]
            self.process.remove(p)
            self.proxy.remove(proxy)
            print(('terminating process', p.pid))
            p.terminate()
            p.kill()


if __name__ == '__main__':
    c = PyroSSH()
    uri = c.nslookup('cmodws30.psfc.mit.edu', 'proxy_efit')
    print(uri)
    print(uri.object)

    proxy = c.open_proxy(uri)
    print(proxy.filename(shot=1120809022, time=1.0, prefix='k'))
    print(proxy.run_minus_5(shot=1120809022, time=1.0))
    c.close_proxy(proxy)



from __future__ import print_function
import sys
import os
import time
import atexit
from signal import SIGTERM


class Daemon:
    """
    A generic daemon class.

    Usage: subclass the Daemon class and override the run() method
    """

    def __init__(self, pidfile, stdin='/dev/null', stdout='/dev/null', stderr='/dev/null'):
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.pidfile = pidfile

    def daemonize(self):
        """
        do the UNIX double-fork magic, see Stevens' "Advanced
        Programming in the UNIX Environment" for details (ISBN 0201563177)
        http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
        """
        try:
            pid = os.fork()
            if pid > 0:
                # exit first parent
                sys.exit(0)
        except OSError as e:
            sys.stderr.write("fork #1 failed: %d (%s)\n" %
                             (e.errno, e.strerror))
            sys.exit(1)

        # decouple from parent environment
        os.chdir("/")
        os.setsid()
        os.umask(0)
        # do second fork
        try:
            pid = os.fork()
            if pid > 0:
                # exit from second parent
                sys.exit(0)
        except OSError as e:
            sys.stderr.write("fork #2 failed: %d (%s)\n" %
                             (e.errno, e.strerror))
            sys.exit(1)

        # redirect standard file descriptors
        sys.stdout.flush()
        sys.stderr.flush()
        si = file(self.stdin, 'r')
        so = file(self.stdout, 'a+')
        se = file(self.stderr, 'a+', 0)
        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())

        # write pidfile
        atexit.register(self.delpid)
        pid = str(os.getpid())
        file(self.pidfile, 'w+').write("%s\n" % pid)
        print((pid, self.pidfile))

    def delpid(self):
        print('Removing pid file')
        os.remove(self.pidfile)

    def start(self):
        """
        Start the daemon
        """
        # Check for a pidfile to see if the daemon already runs
        try:
            pf = file(self.pidfile, 'r')
            pid = int(pf.read().strip())
            pf.close()
        except IOError:
            pid = None

        if pid:
            message = "pidfile %s already exist. Daemon already running?\n"
            sys.stderr.write(message % self.pidfile)
            sys.exit(1)

        # Start the daemon
        self.daemonize()
        self.run()

    def stop(self):
        """
        Stop the daemon
        """
        # Get the pid from the pidfile
        try:
            pf = file(self.pidfile, 'r')
            pid = int(pf.read().strip())

            pf.close()
        except IOError:
            pid = None

        if not pid:
            message = "pidfile %s does not exist. Daemon not running?\n"
            sys.stderr.write(message % self.pidfile)
            return  # not an error in a restart

        # Try killing the daemon process
        try:
            while 1:
                os.kill(pid, SIGTERM)
                time.sleep(0.1)
        except OSError as err:
            err = str(err)
            if err.find("No such process") > 0:
                if os.path.exists(self.pidfile):
                    os.remove(self.pidfile)
            else:
                print(str(err))
                sys.exit(1)

    def restart(self):
        """
        Restart the daemon
        """
        self.stop()
        self.start()

    def run(self):
        """
        You should override this method when you subclass Daemon. It will be called after the process has been
        daemonized by start() or restart().
        """
